#!/usr/bin/env python2.7
""" PreAssembly Report.

See FALCON-pbsmrtpipe/pbfalcon/report_preassembly.py for XML version.
"""
# Copied from
#   http://swarm/files/depot/branches/springfield/S2.3/software/smrtanalysis/bioinformatics/tools/pbreports/pbreports/report/preassembly.py
from __future__ import absolute_import
from __future__ import division
from .FastaReader import FastaReader
from .util.io import syscall
from . import functional
import collections
import glob
import itertools
import logging
import os
import pprint
import re

log = logging.getLogger(__name__)
__version__ = '0.1'

Stats = collections.namedtuple('FastaStats', ['nreads', 'total', 'n50', 'p95'])

# Copied from pbreports/util.py
# We want to avoid a dependency on pbreports b/c it needs matplotlib.
def get_fasta_readlengths(fasta_file):
    """
    Get a sorted list of contig lengths
    :return: (tuple)
    """
    lens = []
    with FastaReader(fasta_file) as f:
        for record in f:
            lens.append(len(record.sequence))
    lens.sort()
    return lens

def get_db_readlengths(fn):
    """Use DBdump on a DAZZ_DB.
    If DBsplit was run, then we see the filtered reads only, since we do not provide '-u' to DBdump.
    """
    call = 'DBdump -h {}'.format(fn)
    return list(functional.parsed_readlengths_from_dbdump_output(syscall(call)))

class FastaContainer(object):

    def __init__(self, nreads, total, file_name):
        self.nreads = nreads
        self.total = total
        self.file_name = file_name

    @staticmethod
    def from_file(file_name):
#        nreads, total = _compute_values(file_name)
        read_lens = get_fasta_readlengths(file_name)
        nreads = len(read_lens)
        total = sum(read_lens)
        return FastaContainer(nreads, total, file_name)

    def __str__(self):
        return "N {n} Total {t} File: {f}".format(n=self.nreads, t=self.total, f=self.file_name)

def _validate_file(file_name):
    if os.path.isfile(file_name):
        return os.path.abspath(file_name)
    else:
        msg = "Unable to find {f}".format(f=file_name)
        log.error(msg)
        raise IOError(msg)

def cutoff_reads(read_lens, min_read_len):
    return [rl for rl in read_lens if rl >= min_read_len]

def read_len_above(read_lens, threshold):
    subtotal = 0
    # Reverse-order calculation is faster.
    for irev, rl in enumerate(reversed(read_lens)):
        subtotal += rl
        if subtotal >= threshold:
            return rl

def percentile(read_lens, p):
    # TODO: Fix this when p=1.0
    return read_lens[int(len(read_lens)*p)]

def stats_from_sorted_readlengths(read_lens):
    nreads = len(read_lens)
    total = sum(read_lens)
    n50 = read_len_above(read_lens, int(total * 0.50))
    p95 = percentile(read_lens, 0.95)
    #alt_n50 = pbreports.util.compute_n50(read_lens)
    #log.info('our n50=%s, pbreports=%s' %(n50, alt_n50)) # Ours is more correct when median is between 2 reads.
    return Stats(nreads=nreads, total=total, n50=n50, p95=p95)

def read_lens_from_fofn(fofn_fn):
    """Return sorted list.
    """
    fns = [fn.strip() for fn in open(fofn_fn) if fn.strip()]
    # get_fasta_readlengths() returns sorted, so sorting the chain is roughly linear.
    return list(sorted(itertools.chain.from_iterable(get_fasta_readlengths(fn) for fn in fns)))

def read_lens_from_db(db_fn):
    """Return sorted read-lengths from a DAZZ_DB.
    """
    return list(sorted(get_db_readlengths(db_fn)))

def abs_filenames(fofn_fn):
    fofn_dir = os.path.dirname(fofn_fn)
    def abs_fn(fn):
        if os.path.isabs(fn):
            return fn
        return os.path.join(fofn_dir, fn)
    fns = [abs_fn(fn.strip()) for fn in open(fofn_fn) if fn.strip()]
    return fns

def metric_fragmentation(preads_fofn):
    # https://jira.pacificbiosciences.com/browse/SAT-105
    #sed -nr 's;>prolog/([0-9]*)[0-9]/.*;\1;p' %s/*.fasta | sort | uniq -c | awk '{print $1}' | sort | uniq -c
    fastas = abs_filenames(preads_fofn)
    call = """perl -e 'while (<>) { if ( m{>[^/]+/(\d+)\d/} ) { $id{$1}++; } }; while (my ($k, $v) = each %%id) { $counts{$v}++; }; while (my ($k, $v) = each %%counts) { print "$v $k\n"; };' %s""" %(' '.join(fastas))
    counts = syscall(call)
    return functional.calc_metric_fragmentation(counts)

def metric_truncation(db, preads_fofn):
    # https://jira.pacificbiosciences.com/browse/SAT-105
    fastas = abs_filenames(preads_fofn)
    call = """perl -e 'while (<>) { if ( m{>[^/]+/0*(\d+)\d/(\d+)_(\d+)} ) { $lengths{(1 + $1)} += ($3 - $2); } }; while (my ($k, $v) = each %%lengths) { print "$k $v\n"; };' %s""" %(' '.join(fastas))
    # The +1 is because of the DBdump readids start at 1, but these start at 0.
    length_pairs_output = syscall(call)
    call = 'DBdump -rh {}'.format(db)
    dbdump_output = syscall(call)
    return functional.calc_metric_truncation(dbdump_output, length_pairs_output)

def stats_dict(stats_raw_reads, stats_seed_reads, stats_corrected_reads, genome_length, length_cutoff,
        fragmentation, truncation):
    """All inputs are paths to fasta files.
    genome_length and length_cutoff can be None.
    """
    log.info('stats for raw reads:       %s' %repr(stats_raw_reads))
    log.info('stats for seed reads:      %s' %repr(stats_seed_reads))
    log.info('stats for corrected reads: %s' %repr(stats_corrected_reads))

    kwds = {}
    genome_length = -1 if not genome_length else genome_length
    kwds['genome_length'] = genome_length
    kwds['length_cutoff'] = 0 if length_cutoff is None else length_cutoff
    kwds['raw_reads'] = stats_raw_reads.nreads
    kwds['raw_bases'] = stats_raw_reads.total
    kwds['raw_mean'] = stats_raw_reads.total / stats_raw_reads.nreads
    kwds['raw_n50'] = stats_raw_reads.n50
    kwds['raw_p95'] = stats_raw_reads.p95
    kwds['raw_coverage'] = stats_raw_reads.total / genome_length
    kwds['seed_reads'] = stats_seed_reads.nreads
    kwds['seed_bases'] = stats_seed_reads.total
    kwds['seed_mean'] = stats_seed_reads.total / stats_seed_reads.nreads
    kwds['seed_n50'] = stats_seed_reads.n50
    kwds['seed_p95'] = stats_seed_reads.p95
    kwds['seed_coverage'] = stats_seed_reads.total / genome_length
    kwds['preassembled_reads'] = stats_corrected_reads.nreads
    kwds['preassembled_bases'] = stats_corrected_reads.total
    kwds['preassembled_mean'] = stats_corrected_reads.total / stats_corrected_reads.nreads
    kwds['preassembled_n50'] = stats_corrected_reads.n50
    kwds['preassembled_p95'] = stats_corrected_reads.p95
    kwds['preassembled_coverage'] = stats_corrected_reads.total / genome_length
    kwds['preassembled_yield'] = stats_corrected_reads.total / stats_seed_reads.total
    kwds['preassembled_seed_fragmentation'] = fragmentation
    kwds['preassembled_seed_truncation'] = truncation
    def round_if_float(v):
        return v if type(v) is not float else round(v, 3)
    result = {k:round_if_float(v) for k,v in kwds.iteritems()}
    return result

# DEPRECATED
def make_dict(
        i_preads_fofn_fn,
        i_raw_reads_fofn_fn,
        genome_length,
        length_cutoff,
    ):
    raw_reads = read_lens_from_fofn(i_raw_reads_fofn_fn)
    stats_raw_reads = stats_from_sorted_readlengths(raw_reads)

    seed_reads = cutoff_reads(raw_reads, length_cutoff)
    stats_seed_reads = stats_from_sorted_readlengths(seed_reads)

    preads = read_lens_from_fofn(i_preads_fofn_fn)
    stats_preads = stats_from_sorted_readlengths(preads)
    report_dict = stats_dict(
            stats_raw_reads=stats_raw_reads,
            stats_seed_reads=stats_seed_reads,
            stats_corrected_reads=stats_preads,
            genome_length=genome_length,
            length_cutoff=length_cutoff,
    )
    return report_dict

def calc_dict(
        i_preads_fofn_fn,
        i_raw_reads_db_fn,
        genome_length,
        length_cutoff,
    ):
    try:
        frag = metric_fragmentation(i_preads_fofn_fn)
    except:
        frag = -1.0
        log.exception('Using arbitrary fragmentation metric: {}'.format(frag))
    try:
        trunc = metric_truncation(i_raw_reads_db_fn, i_preads_fofn_fn)
    except:
        trunc = -1.0
        log.exception('Using arbitrary truncation metric: {}'.format(trunc))

    raw_reads = read_lens_from_db(i_raw_reads_db_fn)
    stats_raw_reads = stats_from_sorted_readlengths(raw_reads)

    seed_reads = cutoff_reads(raw_reads, length_cutoff)
    stats_seed_reads = stats_from_sorted_readlengths(seed_reads)

    preads = read_lens_from_fofn(i_preads_fofn_fn)
    stats_preads = stats_from_sorted_readlengths(preads)
    report_dict = stats_dict(
            stats_raw_reads=stats_raw_reads,
            stats_seed_reads=stats_seed_reads,
            stats_corrected_reads=stats_preads,
            genome_length=genome_length,
            length_cutoff=length_cutoff,
            fragmentation=frag,
            truncation=trunc,
    )
    log.info('Calculated pre-assembly stats:\n{}'.format(
        pprint.pformat(report_dict)))
    return report_dict



from setuptools import setup, Extension, find_packages

setup(
    name = 'pypeflow',
    version='1.0.0',
    author='J. Chin',
    author_email='cschin@infoecho.net',
    license='LICENSE.txt',
    packages = [
        'pypeflow',
        'pwatcher', # a separate package here for convenience, for now
        'pwatcher.mains',
    ],
    package_dir = {'':'.'},
    zip_safe = False,
    install_requires=[
        'networkx >=1.7, <=1.11',
    ],
    entry_points = {'console_scripts': [
            'pwatcher-main=pwatcher.mains.pwatcher:main',
            'pwatcher-pypeflow-example=pwatcher.mains.pypeflow_example:main',
            'heartbeat-wrapper=pwatcher.mains.fs_heartbeat:main',
        ],
    },
    package_data={'pwatcher.mains': ['*.sh']}
)



from falcon_kit import kup, falcon, DWA, get_consensus, get_alignment
from pbcore.io import FastaReader
import numpy as np
import collections
import sys

seqs = []
q_seqs = {}
f = FastaReader(sys.argv[1]) # take one commnad line argument of the input fasta file name

for r in f:
    if len(r.sequence) < 6000:
        continue
    seq = r.sequence.upper()
    seqs.append( (r.name, seq[:500], seq[-500:] ) )
    q_seqs[r.name] = seq


total_index_base = len(seqs) * 1000
print total_index_base
sa_ptr = kup.allocate_seq( total_index_base )
sda_ptr = kup.allocate_seq_addr( total_index_base )
K=14
lk_ptr = kup.allocate_kmer_lookup( 1 << (K * 2) )

start = 0
for r_name, prefix, suffix in seqs:
    kup.add_sequence( start, K, prefix, 500, sda_ptr, sa_ptr, lk_ptr)
    start += 500
    kup.add_sequence( start, K, suffix, 500, sda_ptr, sa_ptr, lk_ptr)
    start += 500
#kup.mask_k_mer(1 << (K * 2), lk_ptr, 256)

kup.mask_k_mer(1 << (K * 2), lk_ptr, 64)

def get_alignment(seq1, seq0):

    K = 8
    lk_ptr = kup.allocate_kmer_lookup( 1 << (K * 2) )
    sa_ptr = kup.allocate_seq( len(seq0) )
    sda_ptr = kup.allocate_seq_addr( len(seq0) )
    kup.add_sequence( 0, K, seq0, len(seq0), sda_ptr, sa_ptr, lk_ptr)

    kmer_match_ptr = kup.find_kmer_pos_for_seq(seq1, len(seq1), K, sda_ptr, lk_ptr)
    kmer_match = kmer_match_ptr[0]
    aln_range = kup.find_best_aln_range(kmer_match_ptr, K, K*5, 50)
    #x,y = zip( * [ (kmer_match.query_pos[i], kmer_match.target_pos[i]) for i in range(kmer_match.count )] )
    kup.free_kmer_match(kmer_match_ptr)
    s1, e1, s2, e2 = aln_range.s1, aln_range.e1, aln_range.s2, aln_range.e2
    if e1 - s1 > 500:
        #s1 = 0 if s1 < 14 else s1 - 14
        #s2 = 0 if s2 < 14 else s2 - 14
        e1 = len(seq1) if e1 >= len(seq1)-2*K else e1 + K*2
        e2 = len(seq0) if e2 >= len(seq0)-2*K else e2 + K*2

        alignment = DWA.align(seq1[s1:e1], e1-s1,
                              seq0[s2:e2], e2-s2,
                              100, 0)
        #print seq1[s1:e1]
        #print seq0[s2:e2]
        #if alignment[0].aln_str_size > 500:

        #aln_str1 = alignment[0].q_aln_str
        #aln_str0 = alignment[0].t_aln_str
        aln_size = alignment[0].aln_str_size
        aln_dist = alignment[0].dist
        aln_q_s = alignment[0].aln_q_s
        aln_q_e = alignment[0].aln_q_e
        aln_t_s = alignment[0].aln_t_s
        aln_t_e = alignment[0].aln_t_e
        assert aln_q_e- aln_q_s <= alignment[0].aln_str_size or aln_t_e- aln_t_s <= alignment[0].aln_str_size
        #print aln_str1
        #print aln_str0

        DWA.free_alignment(alignment)

    kup.free_seq_addr_array(sda_ptr)
    kup.free_seq_array(sa_ptr)
    kup.free_kmer_lookup(lk_ptr)
    if e1 - s1 > 500 and aln_size > 500:
        return s1, s1+aln_q_e-aln_q_s, s2, s2+aln_t_e-aln_t_s, aln_size, aln_dist
    else:
        return None


def get_ovelap_alignment(seq1, seq0):

    K = 8
    lk_ptr = kup.allocate_kmer_lookup( 1 << (K * 2) )
    sa_ptr = kup.allocate_seq( len(seq0) )
    sda_ptr = kup.allocate_seq_addr( len(seq0) )
    kup.add_sequence( 0, K, seq0, len(seq0), sda_ptr, sa_ptr, lk_ptr)

    kmer_match_ptr = kup.find_kmer_pos_for_seq(seq1, len(seq1), K, sda_ptr, lk_ptr)
    kmer_match = kmer_match_ptr[0]
    aln_range = kup.find_best_aln_range(kmer_match_ptr, K, K*5, 50)
    #x,y = zip( * [ (kmer_match.query_pos[i], kmer_match.target_pos[i]) for i in range(kmer_match.count )] )
    kup.free_kmer_match(kmer_match_ptr)
    s1, e1, s0, e0 = aln_range.s1, aln_range.e1, aln_range.s2, aln_range.e2
    len_1 = len(seq1)
    len_0 = len(seq0)
    do_aln = False
    contain_status = "none"
    if e1 - s1 > 500:
        if s1 < 100 and len_1 - e1 < 100:
            do_aln = False
            contain_status = "contains"
        elif s0 < 100 and len_0 - e0 < 100:
            do_aln = False
            contain_status = "contained"
        else:
            do_aln = True
            if s0 < s1:
                s1 -= s0 #assert s1 > 0
                s0 = 0
                e1 = len_1
                e0 = len_1 - s1 if len_1 - s1 < len_0 else len_0
                if e0 == len_0:
                    do_aln = False
                    contain_status = "contained"

            if s1 <= s0:
                s0 -= s1 #assert s1 > 0
                s1 = 0
                e0 = len_0
                e1 = len_0 - s0 if len_0 - s0 < len_1 else len_1
                if e1 == len_1:
                    do_aln = False
                    contain_status = "contains"


        if do_aln:
            alignment = DWA.align(seq1[s1:e1], e1-s1,
                                  seq0[s0:e0], e0-s0,
                                  500, 0)
            #print seq1[s1:e1]
            #print seq0[s2:e2]
            #if alignment[0].aln_str_size > 500:

            #aln_str1 = alignment[0].q_aln_str
            #aln_str0 = alignment[0].t_aln_str
            aln_size = alignment[0].aln_str_size
            aln_dist = alignment[0].dist
            aln_q_s = alignment[0].aln_q_s
            aln_q_e = alignment[0].aln_q_e
            aln_t_s = alignment[0].aln_t_s
            aln_t_e = alignment[0].aln_t_e
            assert aln_q_e- aln_q_s <= alignment[0].aln_str_size or aln_t_e- aln_t_s <= alignment[0].aln_str_size
            #print aln_str1
            #print aln_str0
            if aln_size > 500:
                contain_status = "overlap"
            DWA.free_alignment(alignment)

    kup.free_seq_addr_array(sda_ptr)
    kup.free_seq_array(sa_ptr)
    kup.free_kmer_lookup(lk_ptr)

    if e1 - s1 > 500 and do_aln and aln_size > 500:
        #return s1, s1+aln_q_e-aln_q_s, s2, s2+aln_t_e-aln_t_s, aln_size, aln_dist, x, y
        return s1, s1+aln_q_e-aln_q_s, s0, s0+aln_t_e-aln_t_s, aln_size, aln_dist, contain_status
    else:
        return 0, 0, 0, 0, 0, 0, contain_status

rc_map = dict( zip("ACGTacgtNn-", "TGCAtgcaNn-") )
with open("test_ovlp.dat","w") as f:
    for name, q_seq in q_seqs.items():
        kmer_match_ptr = kup.find_kmer_pos_for_seq(q_seq, len(q_seq), K, sda_ptr, lk_ptr)
        kmer_match = kmer_match_ptr[0]
        count = kmer_match.count
        hit_index = np.array(kmer_match.target_pos[0:count])/500
        kup.free_kmer_match(kmer_match_ptr)

        c = collections.Counter(hit_index)
        s = [c[0] for c in c.items() if c[1] >50]
        #s.sort()
        targets = set()
        for p in s:
            hit_id = seqs[p/2][0]
            if hit_id in targets or hit_id == name:
                continue
            targets.add(hit_id)
            seq1, seq0 = q_seq, q_seqs[hit_id ]
            rtn = get_ovelap_alignment(seq1, seq0)
            #rtn = get_alignment(seq1, seq0)
            if rtn != None:

                s1, e1, s2, e2, aln_size, aln_dist, c_status = rtn
                #print >>f, name, 0, s1, e1, len(seq1), hit_id, 0, s2, e2, len(seq0),  aln_size, aln_dist
                print >>f, hit_id, name, aln_dist - aln_size, "%0.2f" % (100 - 100.0*aln_dist/(aln_size+1)), 0, s2, e2, len(seq0), 0, s1, e1, len(seq1), c_status

        r_q_seq = "".join([rc_map[c] for c in q_seq[::-1]])

        kmer_match_ptr = kup.find_kmer_pos_for_seq(r_q_seq, len(r_q_seq), K, sda_ptr, lk_ptr)
        kmer_match = kmer_match_ptr[0]
        count = kmer_match.count
        hit_index = np.array(kmer_match.target_pos[0:count])/500
        kup.free_kmer_match(kmer_match_ptr)

        c = collections.Counter(hit_index)
        s = [c[0] for c in c.items() if c[1] >50]
        #s.sort()
        targets = set()
        for p in s:
            hit_id = seqs[p/2][0]
            if hit_id in targets or hit_id == name:
                continue
            targets.add(hit_id)
            seq1, seq0 = r_q_seq, q_seqs[hit_id]
            rtn = get_ovelap_alignment(seq1, seq0)
            #rtn = get_alignment(seq1, seq0)
            if rtn != None:
                s1, e1, s2, e2, aln_size, aln_dist, c_status = rtn
                #print >>f, name, 1, s1, e1, len(seq1), hit_id, 0, s2, e2, len(seq0),  aln_size, aln_dist
                print >>f, hit_id, name, aln_dist - aln_size, "%0.2f" % (100 - 100.0*aln_dist/(aln_size+1)), 0, s2, e2, len(seq0), 1, len(seq1) - e1, len(seq1)- s1, len(seq1), c_status




"""Filesystem heartbeat wrapper

Perl might be better for efficiency.
But we will use python for now.

Non-zero status means *this* failed, not the wrapped command.
"""
import argparse
import os
import socket
import sys
import threading
import time

DESCRIPTION = """
We wrap a system call to produce both a heartbeat and an exit-sentinel
in the filesystem.
"""
EPILOG = """
We share stderr/stdout with the command. We log to stderr (for now).
"""
HEARTBEAT_TEMPLATE = '0 {pid} {pgid}\n'
EXIT_TEMPLATE = '{exit_code}'

class _Formatter(argparse.RawDescriptionHelpFormatter, argparse.ArgumentDefaultsHelpFormatter):
    pass
_FORMATTER_CLASS = _Formatter

def parse_args(args):
    parser = argparse.ArgumentParser(
        description=DESCRIPTION,
        epilog=EPILOG,
        formatter_class=_FORMATTER_CLASS,
    )
    parser.add_argument('--rate',
        help='Heartbeat rate, in seconds',
        type=float,
        default=1.0, # TODO: Make this at least 10, maybe 60.
    )
    parser.add_argument('--heartbeat-file',
        help='Path to heartbeat file. The first line will have the format {!r}. The rest are just elapsed time'.format(
            HEARTBEAT_TEMPLATE),
        required=True,
    )
    parser.add_argument('--exit-file',
        help='Path to exit sentinel file. At end, it will have the format {!r}'.format(
            EXIT_TEMPLATE),
        required=True,
    )
    parser.add_argument('--directory',
        help='Directory in which to run COMMAND.',
        default='.',
    )
    parser.add_argument('command',
        help='System call (to be joined by " "). We will block on this and return its result.',
        nargs='+',
        #required=True,
    )
    return parser.parse_args(args)

def log(msg):
    sys.stderr.write(msg)
    sys.stderr.write('\n')
    #sys.stdout.flush() # If we use stdout.

def thread_heartbeat(heartbeat_fn, sleep_s):
    with open(heartbeat_fn, 'w') as ofs:
        pid = os.getpid()
        pgid = os.getpgid(0)
        ofs.write(HEARTBEAT_TEMPLATE.format(
            **locals()))
        elapsed = 0
        ctime = 0
        while True:
            #ctime = time.time()
            ofs.write('{elapsed} {ctime}\n'.format(
                **locals()))
            ofs.flush()
            time.sleep(sleep_s)
            elapsed += 1

def start_heartbeat(heartbeat_fn, sleep_s):
    hb = threading.Thread(target=thread_heartbeat, args=(heartbeat_fn, sleep_s))
    log('alive? {}'.format(
        bool(hb.is_alive())))
    hb.daemon = True
    hb.start()
    return hb

def run(args):
    os.chdir(args.directory)
    heartbeat_fn = os.path.abspath(args.heartbeat_file)
    exit_fn = os.path.abspath(args.exit_file)
    cwd = os.getcwd()
    hostname = socket.getfqdn()
    sleep_s = args.rate
    log("""
cwd:{cwd!r}
hostname={hostname}
heartbeat_fn={heartbeat_fn!r}
exit_fn={exit_fn!r}
sleep_s={sleep_s!r}""".format(
        **locals()))
    if os.path.exists(exit_fn):
        os.remove(exit_fn)
    if os.path.exists(heartbeat_fn):
        os.remove(heartbeat_fn)
    #os.system('touch {}'.format(heartbeat_fn)) # This would be over-written anyway.
    log("before setpgid: pid={} pgid={}".format(os.getpid(), os.getpgid(0)))
    try:
        os.setpgid(0, 0) # This allows the entire tree of procs to be killed.
        log(" after setpgid: pid={} pgid={}".format(
            os.getpid(), os.getpgid(0)))
    except OSError as e:
        log(' Unable to set pgid. Possibly a grid job? Hopefully there will be no dangling processes when killed: {}'.format(
            repr(e)))

    #thread = start_heartbeat(heartbeat_fn, sleep_s)

    #log('alive? {} pid={} pgid={}'.format(
    #    bool(thread.is_alive()), os.getpid(), os.getpgid(0)))

    call = ' '.join(args.command)
    log('In cwd: {}, Blocking call: {!r}'.format(
        os.getcwd(), call))
    rc = os.system(call) # Blocking.

    log(' returned: {!r}'.format(
        rc))

    # Do not delete the heartbeat here. The discoverer of the exit-sentinel will do that,
    # to avoid a race condition.
    #if os.path.exists(heartbeat_fn):
    #    os.remove(heartbeat_fn)

    exit_tmp_fn = exit_fn + '.tmp'
    with open(exit_tmp_fn, 'w') as ofs:
        ofs.write(EXIT_TEMPLATE.format(
            exit_code=rc))
    os.rename(exit_tmp_fn, exit_fn) # atomic
    # sys.exit(rc) # No-one would see this anyway.

def main():
    args = parse_args(sys.argv[1:])
    log(repr(args))
    run(args)

if __name__ == "__main__":
    main()



import subprocess
from pbcore.io import FastaReader

def get_matches(seq0, seq1):
    with open("tmp_seq0.fa","w") as f:
        print >>f, ">seq0"
        print >>f, seq0
    with open("tmp_seq1.fa","w") as f:
        print >>f, ">seq1"
        print >>f, seq1
    mgaps_out=subprocess.check_output("mummer -maxmatch -c -b -l 24 tmp_seq0.fa tmp_seq1.fa | mgaps ", stderr = open("/dev/null"), shell=True)

    matches = []
    cluster = []
    for l in mgaps_out.split("\n"):
        l = l.strip().split()
        if len(l) == 0:
            continue
        if l[0] == ">":
            seq_id = l[1]

            if len(cluster) != 0:
                matches.append(cluster)

            cluster = []
            continue
        if l[0] == "#":
            if len(cluster) != 0:
                matches.append(cluster)
            cluster = []
            continue
        len_ = int(l[2])
        r_s = int(l[0])
        q_s = int(l[1])
        r_e = r_s + len_
        q_e = q_s + len_
        cluster.append( ((r_s, r_e), (q_s, q_e)) )
    if len(cluster) != 0:
        matches.append(cluster)
    return matches


u_edges = {}
with open("./unit_edges.dat") as f:
    for l in f:
        v, w, path, seq = l.strip().split()
        u_edges.setdefault( (v, w), [] )
        u_edges[ (v, w) ].append( (path, seq) )


p_tig_path = {}
a_tig_path = {}
with open("primary_tigs_paths_c") as f:
    for l in f:
        l = l.strip().split()
        id_ = l[0][1:]
        path = l[1:]
        p_tig_path[id_] = path

with open("all_tigs_paths") as f:
    for l in f:
        l = l.strip().split()
        id_ = l[0][1:]
        path = l[1:]
        a_tig_path[id_] = path

p_tig_seqs = {}
for r in FastaReader("primary_tigs_c.fa"):
    p_tig_seqs[r.name] = r.sequence

a_tig_seqs = {}
for r in FastaReader("all_tigs.fa"):
    a_tig_seqs[r.name.split()[0]] = r.sequence

p_tig_to_node_pos = {}
node_pos = []
with open("primary_tigs_node_pos_c") as f:
    for l in f:
        l = l.strip().split()
        p_tig_to_node_pos.setdefault( l[0], [])
        p_tig_to_node_pos[l[0]].append( (l[1], int(l[2])))

duplicate_a_tigs = []
with open("a_nodup.fa","w") as out_f:
    for p_tig_id in p_tig_path:
        main_path = p_tig_path[p_tig_id]
        main_path_nodes = set(main_path[:])
        p_tig_seq = p_tig_seqs[p_tig_id]
        a_node = []
        a_node_range = []
        a_node_range_map = {}
        node_to_pos = dict( p_tig_to_node_pos[p_tig_id] )
        for id_ in a_tig_path:
            if id_[:4] != p_tig_id[:4]:
                continue
            if id_.split("-")[1] == "0000":
                continue

            a_path = a_tig_path[id_]
            if a_path[0] in main_path_nodes and a_path[-1] in main_path_nodes:
                #print p_tig_id, id_, a_path[0], a_path[-1]
                s, e = node_to_pos[a_path[0]], node_to_pos[a_path[-1]]
                p_seq = p_tig_seq[s:e]
                a_seq = a_tig_seqs[id_]
                seq_match = get_matches(p_seq, a_seq)
                if len(seq_match) > 1:
                    print >>out_f, ">"+id_
                    print >>out_f,  a_seq
                    continue
                try:
                    r_s, r_e = seq_match[0][0][0][0], seq_match[0][-1][0][1]
                except:
                    print "XXX", seq_match
                if 1.0* (r_e - r_s) / (e - s) > 98:
                    print >>out_f, ">"+id_
                    print >>out_f, a_seq
                    continue
                duplicate_a_tigs.append(id_)




from .util import (mkdirs, system, touch, run, cd)
import pwatcher.blocking
import pwatcher.fs_based
import pwatcher.network_based
import networkx
import networkx.algorithms.dag #import (topological_sort, is_directed_acyclic_graph)

import hashlib
import json
import logging
import os
import pprint
import random
import sys
import tempfile
import time
import traceback

LOG = logging.getLogger(__name__)

def generate_jobid(node, script_fn):
    # For now, we keep it simple. Just the task.json.
    script_content = open(script_fn).read()
    checksum = hashlib.md5(script_content).hexdigest()
    return 'P' + checksum[0:14]

def endrun(thisnode, status):
    """By convention for now, status is one of:
        'DEAD'
        'UNSUBMITTED' (a pseudo-status defined in the ready-loop of alive())
        'EXIT rc'
    """
    name = status.split()[0]
    if name == 'DEAD':
        LOG.warning(''.join(traceback.format_stack()))
        LOG.error('Task {}\n is DEAD, meaning no HEARTBEAT, but this can be a race-condition. If it was not killed, then restarting might suffice. Otherwise, you might have excessive clock-skew.'.format(thisnode))
        #thisnode.setSatisfied(False) # I think we can leave it unsatisfied. Not sure what happens on retry though.
    elif name == 'UNSUBMITTED':
        LOG.warning(''.join(traceback.format_stack()))
        LOG.error('Task {}\n is UNSUBMITTED, meaning job-submission somehow failed. Possibly a re-start would work. Otherwise, you need to investigate.'.format(thisnode))
        thisnode.setSatisfied(False)
    elif name != 'EXIT':
        raise Exception('Unexpected status {!r}'.format(name))
    else:
        code = int(status.split()[1])
        if 0 != code:
            LOG.error('Task {} failed with exit-code={}'.format(thisnode, code))
            thisnode.setSatisfied(False)
        else:
            thisnode.setSatisfied(True)

class PwatcherTaskQueue(object):
    def enque(self, nodes):
        """This should be an injected dependency.
        Yield any Nodes that could not be submitted.
        Nodes that lacked a script were actually run in-process and are considered
        successful unless they raised an Exception, so they go into the to_report set.
        """
        # Start all "nodes".
        # Note: It is safe to run this block always, but we save a
        # call to pwatcher with 'if ready'.
        LOG.debug('enque nodes:\n%s' %pprint.pformat(nodes))
        jobids = dict()
        #sge_option='-pe smp 8 -q default'
        for node in nodes:
            #node.satisfy() # This would do the job without a process-watcher.
            generated_script_fn = node.execute() # misnomer; this only dumps task.json now
            if not generated_script_fn:
                raise Exception('Missing generated_script_fn for Node {}'.format(node))
                # This task is done.
                endrun(node, 'EXIT 0')
                self.__to_report.append(node)
                continue
                # For now, consider it as "submitted" and finished.
                # (It would throw exception on error.)
            jobid = generate_jobid(node, generated_script_fn)
            self.__known[jobid] = node

            rundir, basename = os.path.split(os.path.abspath(generated_script_fn))
            cmd = '/bin/bash {}'.format(basename)
            LOG.debug('In rundir={!r}, sge_option={!r}, __sge_option={!r}'.format(
                rundir,
                node.pypetask.parameters.get('sge_option'), self.__sge_option))
            sge_option = node.pypetask.parameters.get('sge_option', self.__sge_option)
            job_type = node.pypetask.parameters.get('job_type', None)
            job_queue = node.pypetask.parameters.get('job_queue', None)
            job_nprocs = node.pypetask.parameters.get('job_nprocs', None)
            jobids[jobid] = {
                'cmd': cmd,
                'rundir': rundir,
                # These are optional:
                'job_type': job_type,
                'job_queue': job_queue,
                'job_nprocs': job_nprocs,
                'sge_option': sge_option,
            }
        # Also send the default type and queue-name.
        watcher_args = {
                'jobids': jobids,
                'job_type': self.__job_type,
                'job_queue': self.__job_queue,
        }
        result = self.watcher.run(**watcher_args)
        LOG.debug('Result of watcher.run()={}'.format(repr(result)))
        submitted = result['submitted']
        self.__running.update(submitted)
        #log.info("QQQ ADDED: {}".format(jobid))
        for jobid in (set(jobids.keys()) - set(submitted)):
            yield self.__known[jobid] # TODO: What should be the status for these submission failures?

    def check_done(self):
        """Yield Nodes which have finished since the last check.
        """
        watcher_args = {
            'jobids': list(self.__running),
            'which': 'list',
        }
        q = self.watcher.query(**watcher_args)
        #LOG.debug('In check_done(), result of query:%s' %repr(q))
        for jobid, status in q['jobids'].iteritems():
            if status.startswith('EXIT') or status.startswith('DEAD'):
                self.__running.remove(jobid)
                node = self.__known[jobid]
                self.__to_report.append(node)
                try:
                    endrun(node, status)
                except Exception as e:
                    msg = 'Failed to clean-up FakeThread: jobid={} status={}'.format(jobid, repr(status))
                    LOG.exception(msg)
                    raise
        to_report = list(self.__to_report)
        self.__to_report = list()
        return to_report

    def terminate(self):
        watcher_args = {
            'jobids': list(self.__running),
            'which': 'known',
        }
        q = self.watcher.delete(**watcher_args)
        LOG.debug('In notifyTerminate(), result of delete:%s' %repr(q))

    def __init__(self, watcher, job_type='local', job_queue=None, sge_option=None):
        self.watcher = watcher
        self.__job_type = job_type
        self.__job_queue = job_queue
        self.__sge_option = sge_option
        self.__running = set() # jobids
        self.__known = dict() # jobid -> Node
        self.__to_report = list() # Nodes

def get_unsatisfied_subgraph(g):
    unsatg = networkx.DiGraph(g)
    for n in g.nodes():
        if n.satisfied():
            unsatg.remove_node(n)
    return unsatg
def find_all_ancestors(g):
    ancestors = dict()
    def get_ancestors(node):
        """memoized"""
        if node not in ancestors:
            myancestors = set()
            for pred in g.predecessors_iter(node):
                myancestors.update(get_ancestors(pred))
            ancestors[node] = myancestors
        return ancestors[node]
    for node in g.nodes():
        get_ancestors(node)
    return ancestors
def find_next_ready_and_remove(g, node):
    """Given a recently satisfied node,
    return any successors with in_degree 1.
    Then remove node from g immediately.
    """
    ready = set()
    for n in g.successors_iter(node):
        if 1 == g.in_degree(n):
            ready.add(n)
    g.remove_node(node)
    return ready
def find_all_roots(g):
    """Find all nodes in g which have no predecessors.
    """
    roots = set()
    for node in g:
        if 0 == g.in_degree(node):
            roots.add(node)
    return roots

class Workflow(object):
    def _create_node(self, pypetask):
        """Given a PypeTask, return a Node for our Workflow graph.
        Recursively create more nodes based on 'pypetask.inputs',
        record them as 'node.needs', and update the global pypetask2node table.
        """
        needs = set()
        use_tmpdir = self.use_tmpdir # TODO(CD): Let pypetask.use_tmpdir override this.
        node = PypeNode(pypetask.name, pypetask.wdir, pypetask, needs, use_tmpdir) #, pypetask.generated_script_fn)
        self.pypetask2node[pypetask] = node
        for key, plf in pypetask.inputs.iteritems():
            if plf.producer is None:
                continue
            if plf.producer not in self.pypetask2node:
                self._create_node(plf.producer)
            needed_node = self.pypetask2node[plf.producer]
            needs.add(needed_node)
        LOG.debug('New {!r} needs {!r}'.format(node, needs))
        return node
    def addTask(self, pypetask):
        node = self._create_node(pypetask)
        sentinel_done_fn = node.sentinel_done_fn()
        if sentinel_done_fn in self.sentinels:
            msg = 'FOUND sentinel {!r} twice: {!r} ({!r}) and {!r}\nNote: Each task needs its own sentinel (and preferably its own run-dir).'.format(sentinel_done_fn, node, pypetask, self.sentinels[sentinel_done_fn])
            raise Exception(msg)
        self.sentinels[sentinel_done_fn] = node
        self.graph.add_node(node)
        for need in node.needs:
            #print "Need:", need, node
            self.graph.add_edge(need, node)
    def addTasks(self, tlist):
        for t in tlist:
            self.addTask(t)
    def refreshTargets(self, targets=None, updateFreq=10, exitOnFailure=True):
        try:
            self._refreshTargets(updateFreq, exitOnFailure)
        except:
            self.tq.terminate()
            raise
    def _refreshTargets(self, updateFreq, exitOnFailure):
        """Raise Exception (eventually) on any failure.
        - updateFreq (seconds) is really a max; we climb toward it gradually, and we reset when things change.
        - exitOnFailure=False would allow us to keep running (for a while) when parallel jobs fail.
        - self.max_jobs: the max number of concurrently running jobs
          - possibly this should be the number of cpus in use, but for now it is qsub jobs, for simplicity.
        """
        # Note: With the 'blocking' pwatcher, we will have 1 thread per live qsub job.
        # If/when that becomes a problem, please re-write the pwatcher as Go or Nim.
        # This loop is single-threaded. If we ignore max_jobs,
        # then we would send large queries for no reason, but that is not really a big deal.
        # The Python 'blocking' pwatcher is the real reason to limit jobs, for now.
        assert networkx.algorithms.dag.is_directed_acyclic_graph(self.graph)
        failures = 0
        unsatg = get_unsatisfied_subgraph(self.graph)
        ready = find_all_roots(unsatg)
        submitted = set()
        init_sleep_time = 0.1
        sleep_time = init_sleep_time
        LOG.info('Num unsatisfied: {}, graph: {}'.format(len(unsatg), len(self.graph)))
        while ready or submitted:
            # Nodes cannot be in ready or submitted unless they are also in unsatg.
            to_submit = set()
            while ready and (self.max_jobs > len(submitted) + len(to_submit)):
                node = ready.pop()
                to_submit.add(node)
                LOG.info('About to submit: {!r}'.format(node))
            if to_submit:
                unsubmitted = set(self.tq.enque(to_submit)) # In theory, this traps exceptions.
                if unsubmitted:
                    LOG.error('Failed to enqueue {} of {} jobs: {!r}'.format(
                        len(unsubmitted), len(to_submit), unsubmitted))
                    #ready.update(unsubmitted) # Resubmit only in pwatcher, if at all.
                submitted.update(to_submit - unsubmitted)
            LOG.debug('N in queue: {} (max_jobs={})'.format(len(submitted), self.max_jobs))
            recently_done = set(self.tq.check_done())
            if not recently_done:
                if not submitted:
                    LOG.warning('Nothing is happening, and we had {} failures. Should we quit? Instead, we will just sleep.'.format(failures))
                    #break
                LOG.info('sleep {}s'.format(sleep_time))
                time.sleep(sleep_time)
                sleep_time = sleep_time + 0.1 if (sleep_time < updateFreq) else updateFreq
                continue
            LOG.debug('recently_done: {!r}'.format([(rd, rd.satisfied()) for rd in recently_done]))
            LOG.debug('Num done in this iteration: {}'.format(len(recently_done)))
            sleep_time = init_sleep_time
            submitted -= recently_done
            recently_satisfied = set(n for n in recently_done if n.satisfied())
            recently_done -= recently_satisfied
            #LOG.debug('Now N recently_done: {}'.format(len(recently_done)))
            LOG.info('recently_satisfied: {!r}'.format(recently_satisfied))
            LOG.info('Num satisfied in this iteration: {}'.format(len(recently_satisfied)))
            for node in recently_satisfied:
                ready.update(find_next_ready_and_remove(unsatg, node))
            LOG.info('Num still unsatisfied: {}'.format(len(unsatg)))
            if recently_done:
                msg = 'Some tasks are recently_done but not satisfied: {!r}'.format(recently_done)
                LOG.error(msg)
                LOG.error('ready: {!r}\nsubmitted: {!r}'.format(ready, submitted))
                failures += len(recently_done)
                if exitOnFailure:
                    raise Exception(msg)
        assert not submitted
        assert not ready
        if failures:
            raise Exception('We had {} failures. {} tasks remain unsatisfied.'.format(
                failures, len(unsatg)))

    def __init__(self, watcher, job_type, job_queue, max_jobs, use_tmpdir, squash, sge_option=''):
        self.graph = networkx.DiGraph()
        self.tq = PwatcherTaskQueue(watcher=watcher, job_type=job_type, job_queue=job_queue, sge_option=sge_option) # TODO: Inject this.
        assert max_jobs > 0, 'max_jobs needs to be set. If you use the "blocking" process-watcher, it is also the number of threads.'
        self.max_jobs = max_jobs
        self.sentinels = dict() # sentinel_done_fn -> Node
        self.pypetask2node = dict()
        self.use_tmpdir = use_tmpdir
        self.squash = squash # This really should depend on each Task, but for now a global is fine.
        # For small genomes, serial tasks should always be squashed.

class NodeBase(object):
    """Graph node.
    Can be satisfied on demand, but usually we call execute() and later run the script.
    """
    def setSatisfied(self, s):
        self.__satisfied = s
    def workdir(self):
        return self.wdir
    def script_fn(self):
        return os.path.join(self.workdir(), 'run.sh')
    def sentinel_done_fn(self):
        return self.script_fn() + '.done'
    def satisfied(self):
        """Not just done, but successful.
        If we see the sentinel file, then we memoize True.
        But if we finished a distributed job with exit-code 0, we do not need
        to wait for the sentinel; we know we had success.
        """
        #LOG.debug('Checking satisfied({!r}) for sentinel {!r}'.format(self, self.sentinel_done_fn()))
        if self.__satisfied is not None:
            return self.__satisfied
        if os.path.exists(self.sentinel_done_fn()):
            self.setSatisfied(True)
        return self.__satisfied == True
    def satisfy(self):
        if self.__satisfied:
            return
        # Technically, we might need to cd into wdir first.
        script_fn = self.execute()
        if script_fn:
            run(script_fn)
        self.__satisfied = True
    def execute(self):
        try:
            actual_script_fn = self.generate_script()
        except Exception:
            LOG.exception('Failed generate_script() for {!r}'.format(self))
            raise
        sentinel_done_fn = self.sentinel_done_fn()
        wdir = self.workdir()
        rel_actual_script_fn = os.path.relpath(actual_script_fn, wdir)
        wrapper = """#!/bin/sh
set -vex
cd {wdir}
bash {rel_actual_script_fn}
touch {sentinel_done_fn}
""".format(**locals())
        wrapper_fn = self.script_fn()
        #mkdirs(os.path.dirname(wrapper_fn))
        with open(wrapper_fn, 'w') as ofs:
            ofs.write(wrapper)
        return wrapper_fn
    def generate_script(self):
        raise NotImplementedError(repr(self))
    def __repr__(self):
        return 'Node({})'.format(self.name)
    def __init__(self, name, wdir, needs):
        self.__satisfied = None  # satisfiable
        self.name = name
        self.wdir = wdir
        self.needs = needs
class PypeNode(NodeBase):
    """
    We will clean this up later. For now, it is pretty tightly coupled to PypeTask.
    """
    def generate_script(self):
        wdir = self.wdir
        mkdirs(wdir)
        pt = self.pypetask
        assert pt.wdir == self.wdir
        inputs = {k:v.path for k,v in pt.inputs.items()}
        outputs = {k:os.path.relpath(v.path, wdir) for k,v in pt.outputs.items()}
        for v in outputs.values():
            assert not os.path.isabs(v), '{!r} is not relative'.format(v)
        task_desc = {
                'inputs': inputs,
                'outputs': outputs,
                'parameters': pt.parameters,
                'python_function': pt.func_name,
        }
        task_content = json.dumps(task_desc, sort_keys=True, indent=4, separators=(',', ': '))
        task_json_fn = os.path.join(wdir, 'task.json')
        open(task_json_fn, 'w').write(task_content)
        python = 'python2.7' # sys.executable fails sometimes because of binwrapper: SE-152
        tmpdir_flag = '--tmpdir {}'.format(self.use_tmpdir) if self.use_tmpdir else ''
        cmd = '{} -m pypeflow.do_task {} {}'.format(python, tmpdir_flag, task_json_fn)
        script_content = """#!/bin/bash
hostname
env | sort
pwd
time {cmd}
""".format(**locals())
        script_fn = os.path.join(wdir, 'task.sh')
        open(script_fn, 'w').write(script_content)
        return script_fn
    def old_generate_script(self):
        pt = self.pypetask
        func = pt.func
        func(pt) # Run the function! Probably just generate a script.
        generated_script_fn = getattr(pt, 'generated_script_fn', None) # by convention
        try:
            # Maybe we should require a script always.
            generated_script_fn = pt.generated_script_fn
        except Exception:
            LOG.exception('name={!r} URL={!r}'.format(pt.name, pt.URL))
            raise
        return generated_script_fn
    def __init__(self, name, wdir, pypetask, needs, use_tmpdir): #, script_fn):
        super(PypeNode, self).__init__(name, wdir, needs)
        self.pypetask = pypetask
        self.use_tmpdir = use_tmpdir

# This global exists only because we continue to support the old PypeTask style,
# where a PypeLocalFile does not know the PypeTask which produces it.
# (This also allows us to specify PypeTasks out of order, fwiw.)
# Someday, we might require PypeTasks to depend on outputs of other PypeTasks explicitly;
# then we can drop this dict.
PRODUCERS = dict()

def findPypeLocalFile(path):
    """Look-up based on tail dirname.
    Do not call this for paths relative to their work-dirs.
    """
    assert os.path.isabs(path)
    basename = os.path.basename(path)
    basedir = os.path.relpath(os.path.dirname(path))
    producer = PRODUCERS.get(basedir)
    if producer is None:
        msg = 'No producer PypeTask for basedir {!r} from path {!r} -- Pure input?'.format(
            basedir, path)
        LOG.debug(msg)
        return PypeLocalFile(path, None)
    siblings = producer.outputs
    for plf in siblings.values():
        sibling_basename = os.path.basename(plf.path)
        if sibling_basename == basename:
            return plf
    raise Exception('Failed to find a PypeLocalFile for {!r} among outputs of {!r}'.format(
        path, producer))
def find_work_dir(paths):
    """Return absolute path to directory of all these paths.
    Must be same for all.
    """
    dirnames = set(os.path.dirname(os.path.normpath(p)) for p in paths)
    if len(dirnames) != 1:
        raise Exception('Cannot find work-dir for paths in multiple (or zero) dirs: {} in {}'.format(
            pprint.pformat(paths), pprint.pformat(dirnames)))
    d = dirnames.pop()
    return os.path.abspath(d)
class PypeLocalFile(object):
    def __repr__(self):
        if self.producer:
            path = os.path.relpath(self.path, self.producer.wdir)
            return 'PLF({!r}, {!r})'.format(path, os.path.relpath(self.producer.wdir))
        else:
            path = os.path.relpath(self.path)
            wdir = None
            return 'PLF({!r}, {!r})'.format(path, None)
    def __init__(self, path, producer):
        self.path = os.path.abspath(path)
        self.producer = producer
def makePypeLocalFile(p, producer=None):
    return PypeLocalFile(p, producer)
def fn(p):
    """This must be run in the top run-dir.
    All task funcs are executed there.
    """
    if isinstance(p, PypeLocalFile):
        p = p.path
    return os.path.abspath(p)
def only_path(p):
    if isinstance(p, PypeLocalFile):
        return p.path
    else:
        return p
def PypeTask(inputs, outputs, parameters=None, wdir=None):
    """A slightly messy factory because we want to support both strings and PypeLocalFiles, for now.
    This can alter dict values in inputs/outputs if they were not already PypeLocalFiles.
    """
    if wdir is None:
        #wdir = parameters.get('wdir', name) # One of these must be a string!
        wdir = find_work_dir([only_path(v) for v in outputs.values()])
        # Since we derived wdir from outputs, we need to ensure that they
        # are not relative to wdir.
        for k,v in outputs.items():
            if not isinstance(v, PypeLocalFile) and not os.path.isabs(v):
                outputs[k] = os.path.relpath(v, wdir)
    if not os.path.isabs(wdir):
        wdir = os.path.abspath(wdir)
    this = _PypeTask(inputs, outputs, wdir, parameters)
    #basedir = os.path.basename(wdir)
    basedir = this.name
    if basedir in PRODUCERS:
        raise Exception('Basedir {!r} already used for {!r}. Cannot create new PypeTask {!r}.'.format(
            basedir, PRODUCERS[basedir], this))
    LOG.debug('Added PRODUCERS[{!r}] = {!r}'.format(basedir, this))
    PRODUCERS[basedir] = this
    this.canonicalize()
    this.assert_canonical()
    LOG.debug('Built {!r}'.format(this))
    return this
class _PypeTask(object):
    """Adaptor from old PypeTask API.
    """
    def canonicalize(self):
        for key, val in self.outputs.items():
            if not isinstance(val, PypeLocalFile):
                # If relative, then it is relative to our wdir.
                #LOG.warning('Making PLF: {!r} {!r}'.format(val, self))
                if not os.path.isabs(val):
                    val = os.path.join(self.wdir, val)
                #LOG.warning('Muking PLF: {!r} {!r}'.format(val, self))
                val = PypeLocalFile(val, self)
                self.outputs[key] = val
            else:
                val.producer = self
        for key, val in self.inputs.items():
            if not isinstance(val, PypeLocalFile):
                assert os.path.isabs(val), 'Inputs cannot be relative at self point: {!r} in {!r}'.format(val, self)
                self.inputs[key] = findPypeLocalFile(val)
    def assert_canonical(self):
        # Output values will be updated after PypeTask init, so refer back to self as producer.
        for k,v in self.inputs.iteritems():
            assert os.path.isabs(v.path), 'For {!r}, input {!r} is not absolute'.format(self.wdir, v)
        for k,v in self.outputs.iteritems():
            assert os.path.isabs(v.path), 'For {!r}, output {!r} is not absolute'.format(self.wdir, v)
        common = set(self.inputs.keys()) & set(self.outputs.keys())
        assert (not common), 'Keys in both inputs and outputs of PypeTask({}): {!r}'.format(wdir, common)
    def __call__(self, func):
        self.func = func
        self.func_name = '{}.{}'.format(func.__module__, func.__name__)
        return self
    def __repr__(self):
        return 'PypeTask({!r}, {!r}, {!r}, {!r})'.format(self.name, self.wdir, pprint.pformat(self.outputs), pprint.pformat(self.inputs))
    def __init__(self, inputs, outputs, wdir, parameters=None):
        if parameters is None:
            parameters = {}
        name = os.path.relpath(wdir)
        URL = 'task://localhost/{}'.format(name)
        self.inputs = inputs
        self.outputs = outputs
        self.parameters = dict(parameters) # Always copy this, so caller can re-use, for convenience.
        self.wdir = wdir
        self.name = name
        self.URL = URL
        #for key, bn in inputs.iteritems():
        #    setattr(self, key, os.path.abspath(bn))
        #for key, bn in outputs.iteritems():
        #    setattr(self, key, os.path.abspath(bn))
        assert self.wdir, 'No wdir for {!r} {!r}'.format(self.name, self.URL)
        LOG.debug('Created {!r}'.format(self))

class DummyPypeTask(_PypeTask):
    def __init__(self):
        super(DummyPypeTask, self).__init__({}, {}, {}, wdir='/')

#ROOT_PYPE_TASK = DummyPypeTask()

MyFakePypeThreadTaskBase = None  # just a symbol, not really used

# Here is the main factory.
def PypeProcWatcherWorkflow(
        URL = None,
        job_type='local',
        job_queue='UNSPECIFIED_QUEUE',
        watcher_type='fs_based',
        watcher_directory='mypwatcher',
        max_jobs = 24, # must be > 0, but not too high
        sge_option = None,
        use_tmpdir = None,
        squash = False,
        **attributes):
    """Factory for the workflow.
    """
    if watcher_type == 'blocking':
        pwatcher_impl = pwatcher.blocking
    elif watcher_type == 'network_based':
        pwatcher_impl = pwatcher.network_based
    else:
        pwatcher_impl = pwatcher.fs_based
    LOG.warning('In simple_pwatcher_bridge, pwatcher_impl={!r}'.format(pwatcher_impl))
    LOG.info('In simple_pwatcher_bridge, pwatcher_impl={!r}'.format(pwatcher_impl))
    watcher = pwatcher_impl.get_process_watcher(watcher_directory)
    if use_tmpdir:
        try:
            if not os.path.isabs(use_tmpdir):
                use_tmpdir = os.path.abspath(use_tmpdir)
        except (TypeError, AttributeError):
            use_tmpdir = tempfile.gettempdir()
    LOG.info('job_type={!r}, job_queue={!r}, sge_option={!r}, use_tmpdir={!r}, squash={!r}'.format(
        job_type, job_queue, sge_option, use_tmpdir, squash,
    ))
    return Workflow(watcher,
            job_type=job_type, job_queue=job_queue, max_jobs=max_jobs, sge_option=sge_option, use_tmpdir=use_tmpdir,
            squash=squash,
    )
    #th = MyPypeFakeThreadsHandler('mypwatcher', job_type, job_queue)
    #mq = MyMessageQueue()
    #se = MyFakeShutdownEvent() # TODO: Save pwatcher state on ShutdownEvent. (Not needed for blocking pwatcher. Mildly useful for fs_based.)
    #return pypeflow.controller._PypeConcurrentWorkflow(URL=URL, thread_handler=th, messageQueue=mq, shutdown_event=se,
    #        attributes=attributes)

__all__ = ['PypeProcWatcherWorkflow', 'fn', 'makePypeLocalFile', 'MyFakePypeThreadTaskBase', 'PypeTask']



from nose import SkipTest
from nose.tools import assert_equal
import pypeflow.task
import pypeflow.data
import pypeflow.controller

class TestPypeNode:
    def test___init__(self):
        # pype_node = PypeNode(obj)
        raise SkipTest # TODO: implement your test here

    def test_addAnInNode(self):
        # pype_node = PypeNode(obj)
        # assert_equal(expected, pype_node.addAnInNode(obj))
        raise SkipTest # TODO: implement your test here

    def test_addAnOutNode(self):
        # pype_node = PypeNode(obj)
        # assert_equal(expected, pype_node.addAnOutNode(obj))
        raise SkipTest # TODO: implement your test here

    def test_depth(self):
        # pype_node = PypeNode(obj)
        # assert_equal(expected, pype_node.depth())
        raise SkipTest # TODO: implement your test here

    def test_inDegree(self):
        # pype_node = PypeNode(obj)
        # assert_equal(expected, pype_node.inDegree())
        raise SkipTest # TODO: implement your test here

    def test_outDegree(self):
        # pype_node = PypeNode(obj)
        # assert_equal(expected, pype_node.outDegree())
        raise SkipTest # TODO: implement your test here

    def test_removeAnInNode(self):
        # pype_node = PypeNode(obj)
        # assert_equal(expected, pype_node.removeAnInNode(obj))
        raise SkipTest # TODO: implement your test here

    def test_removeAnOutNode(self):
        # pype_node = PypeNode(obj)
        # assert_equal(expected, pype_node.removeAnOutNode(obj))
        raise SkipTest # TODO: implement your test here

class TestPypeGraph:
    def test___getitem__(self):
        # pype_graph = PypeGraph(RDFGraph, subGraphNodes)
        # assert_equal(expected, pype_graph.__getitem__(url))
        raise SkipTest # TODO: implement your test here

    def test___init__(self):
        # pype_graph = PypeGraph(RDFGraph, subGraphNodes)
        raise SkipTest # TODO: implement your test here

    def test_tSort(self):
        # pype_graph = PypeGraph(RDFGraph, subGraphNodes)
        # assert_equal(expected, pype_graph.tSort())
        raise SkipTest # TODO: implement your test here

class TestPypeWorkflow:
    def test___init__(self):
        # pype_workflow = PypeWorkflow(URL, **attributes)
        raise SkipTest # TODO: implement your test here

    def test_addObject(self):
        # pype_workflow = PypeWorkflow(URL, **attributes)
        # assert_equal(expected, pype_workflow.addObject(obj))
        raise SkipTest # TODO: implement your test here

    def test_addObjects(self):
        # pype_workflow = PypeWorkflow(URL, **attributes)
        # assert_equal(expected, pype_workflow.addObjects(objs))
        raise SkipTest # TODO: implement your test here

    def test_addTask(self):
        # pype_workflow = PypeWorkflow(URL, **attributes)
        # assert_equal(expected, pype_workflow.addTask(taskObj))
        raise SkipTest # TODO: implement your test here

    def test_addTasks(self):
        # pype_workflow = PypeWorkflow(URL, **attributes)
        # assert_equal(expected, pype_workflow.addTasks(taskObjs))
        raise SkipTest # TODO: implement your test here

    def test_dataObjects(self):
        # pype_workflow = PypeWorkflow(URL, **attributes)
        # assert_equal(expected, pype_workflow.dataObjects())
        raise SkipTest # TODO: implement your test here

    def test_graphvizDot(self):
        # pype_workflow = PypeWorkflow(URL, **attributes)
        # assert_equal(expected, pype_workflow.graphvizDot())
        raise SkipTest # TODO: implement your test here

    def test_graphvizShortNameDot(self):
        # pype_workflow = PypeWorkflow(URL, **attributes)
        # assert_equal(expected, pype_workflow.graphvizShortNameDot())
        raise SkipTest # TODO: implement your test here

    def test_makeFileStr(self):
        # pype_workflow = PypeWorkflow(URL, **attributes)
        # assert_equal(expected, pype_workflow.makeFileStr())
        raise SkipTest # TODO: implement your test here

    def test_refreshTargets(self):
        # pype_workflow = PypeWorkflow(URL, **attributes)
        # assert_equal(expected, pype_workflow.refreshTargets(objs, callback))
        raise SkipTest # TODO: implement your test here

    def test_removeObjects(self):
        # pype_workflow = PypeWorkflow(URL, **attributes)
        # assert_equal(expected, pype_workflow.removeObjects(objs))
        raise SkipTest # TODO: implement your test here

    def test_removeTask(self):
        # pype_workflow = PypeWorkflow(URL, **attributes)
        # assert_equal(expected, pype_workflow.removeTask(taskObj))
        raise SkipTest # TODO: implement your test here

    def test_removeTasks(self):
        # pype_workflow = PypeWorkflow(URL, **attributes)
        # assert_equal(expected, pype_workflow.removeTasks(taskObjs))
        raise SkipTest # TODO: implement your test here

    def test_setReferenceRDFGraph(self):
        # pype_workflow = PypeWorkflow(URL, **attributes)
        # assert_equal(expected, pype_workflow.setReferenceRDFGraph(fn))
        raise SkipTest # TODO: implement your test here

    def test_tasks(self):
        # pype_workflow = PypeWorkflow(URL, **attributes)
        # assert_equal(expected, pype_workflow.tasks())
        raise SkipTest # TODO: implement your test here

    def test_scatterTask(self):
        
        import os
        os.system("rm -rf /tmp/pypetest/*")
        nChunk = 3 

        infileObj0 =\
        pypeflow.data.PypeLocalFile(
                      "file://localhost/tmp/pypetest/test_in_0.txt")
        with open(infileObj0.localFileName,"w") as f:
            f.write("prefix4:")

        infileObj4 =\
        pypeflow.data.PypeSplittableLocalFile(
                      "splittablefile://localhost/tmp/pypetest/test_in_4.txt", 
                      nChunk = nChunk)

        with open(infileObj4.localFileName, "w") as f:
            for i in range(nChunk):
                f.write("file%02d\n" % i)

        def scatter(*argv, **kwargv):
            outputObjs = sorted( kwargv["outputDataObjs"].items() )
            nOut = len(outputObjs)
            outputObjs = [ (o[0], o[1], open(o[1].localFileName, "w")) for o in outputObjs]
            with open(kwargv["inputDataObjs"]["completeFile"].localFileName,"r") as f:
                i = 0
                for l in f:
                    outf = outputObjs[i % nOut][2]
                    outf.write(l)
                    i += 1
            for o in outputObjs:
                o[2].close()

        PypeShellTask = pypeflow.task.PypeShellTask
        PypeTask = pypeflow.task.PypeTask
        PypeTaskBase = pypeflow.task.PypeTaskBase
        infileObj4.setScatterTask(PypeTask, PypeTaskBase, scatter)
        
        def gather(*argv, **kwargv):
            inputObjs = sorted( kwargv["inputDataObjs"].items() )
            with open(kwargv["outputDataObjs"]["completeFile"].localFileName,"w") as outf:
                for k, subfile in inputObjs:
                    f = open(subfile.localFileName)
                    outf.write(f.read())
                    f.close()

        outfileObj4 =\
        pypeflow.data.PypeSplittableLocalFile(
                      "splittablefile://localhost/tmp/pypetest/test_out_4.txt", 
                      nChunk = nChunk)

        outfileObj4.setGatherTask(PypeTask, PypeTaskBase, gather)

        PypeScatteredTasks = pypeflow.task.PypeScatteredTasks

        @PypeScatteredTasks( inputDataObjs = {"inf":infileObj4, "prefix":infileObj0},
                             outputDataObjs = {"outf":outfileObj4},
                             URL="tasks://test_fun_4")
        def test_fun_4(*argv, **kwargv):
            chunk_id = kwargv["chunk_id"]
            self = test_fun_4[chunk_id]

            assert self.inf._path == "/tmp/pypetest/%03d_test_in_4.txt" % chunk_id
            with open( self.prefix.localFileName, "r") as f:
                prefix = f.read()

            with open( self.outf._path, "w") as f:
                in_f = open(self.inf.localFileName,"r")
                f.write(prefix + in_f.read())
                in_f.close()
            return self.inf._path

        outfileObj5 =\
        pypeflow.data.PypeSplittableLocalFile(
                      "splittablefile://localhost/tmp/pypetest/test_out_5.txt", 
                      nChunk = nChunk)
        outfileObj5.setGatherTask(PypeTask, PypeTaskBase, gather)

        @PypeScatteredTasks( inputDataObjs = {"inf":infileObj4, "prefix":infileObj0},
                             outputDataObjs = {"outf":outfileObj5},
                             URL="tasks://test_fun_5")
        def test_fun_5(*argv, **kwargv):
            chunk_id = kwargv["chunk_id"]
            self = test_fun_5[chunk_id]

            assert self.inf._path == "/tmp/pypetest/%03d_test_in_4.txt" % chunk_id
            with open( self.prefix.localFileName, "r") as f:
                prefix = f.read()

            with open( self.outf._path, "w") as f:
                in_f = open(self.inf.localFileName,"r")
                f.write(prefix +"2:"+ in_f.read())
                in_f.close()
            return self.inf._path
        assert len(test_fun_4.getTasks()) == nChunk 

        wf = pypeflow.controller.PypeWorkflow()
        wf.addTasks( [test_fun_4, test_fun_5] )
        print wf.graphvizDot
        wf.refreshTargets( [outfileObj4, outfileObj5] )
    
class TestPypeThreadWorkflow:
    def test___init__(self):
        # pype_thread_workflow = PypeThreadWorkflow(URL, **attributes)
        raise SkipTest # TODO: implement your test here

    def test_addTasks(self):
        # pype_thread_workflow = PypeThreadWorkflow(URL, **attributes)
        # assert_equal(expected, pype_thread_workflow.addTasks(taskObjs))
        raise SkipTest # TODO: implement your test here

    def test_refreshTargets(self):
        # pype_thread_workflow = PypeThreadWorkflow(URL, **attributes)
        # assert_equal(expected, pype_thread_workflow.refreshTargets(objs, callback, updateFreq, exitOnFailure))
        raise SkipTest # TODO: implement your test here

    def test_setNumThreadAllowed(self):
        # pype_thread_workflow = PypeThreadWorkflow(URL, **attributes)
        # assert_equal(expected, pype_thread_workflow.setNumThreadAllowed(nT, nS))
        raise SkipTest # TODO: implement your test here

    def test_mutableDataObjects(self):

        infileObj =\
        pypeflow.data.PypeLocalFile("file://localhost/tmp/pypetest/test_for_shared_output_in.txt")

        outfileObj =\
        pypeflow.data.PypeLocalFile("file://localhost/tmp/pypetest/test_for_shared_output_out.txt")

        out1 =\
        pypeflow.data.PypeLocalFile("file://localhost/tmp/pypetest/test_for_shared_output_out1.txt")

        out2 =\
        pypeflow.data.PypeLocalFile("file://localhost/tmp/pypetest/test_for_shared_output_out2.txt")

        out3 =\
        pypeflow.data.PypeLocalFile("file://localhost/tmp/pypetest/test_for_shared_output_out3.txt")

        import os
        os.system("rm -rf /tmp/pypetest/*")

        with open(infileObj.localFileName,"w") as f:
            f.write("test")

        PypeThreadWorkflow = pypeflow.controller.PypeThreadWorkflow
        PypeThreadTaskBase = pypeflow.controller.PypeThreadTaskBase
        PypeTask = pypeflow.task.PypeTask
        wf = PypeThreadWorkflow()

        @PypeTask(mutableDataObjs={"out":outfileObj},
                  outputDataObjs={"out1":out1},
                  inputDataObjs={"in":infileObj},
                  TaskType=PypeThreadTaskBase)
        def task1(task):
            with open(task.out.localFileName, "a") as f:
                print >>f, "written by task1"
            with open(task.out1.localFileName, "w") as f:
                print >>f, "written by task1"

        @PypeTask(mutableDataObjs={"out":outfileObj},
                  outputDataObjs={"out2":out2},
                  inputDataObjs={"in":infileObj},
                  TaskType=PypeThreadTaskBase)
        def task2(task):
            with open(task.out.localFileName, "a") as f:
                print >>f, "written by task2"
            with open(task.out2.localFileName, "w") as f:
                print >>f, "written by task2"

        @PypeTask(mutableDataObjs={"out":outfileObj},
                  outputDataObjs={"out3":out3},
                  inputDataObjs={"in":infileObj},
                  TaskType=PypeThreadTaskBase)
        def task3(task):
            with open(task.out.localFileName, "a") as f:
                print >>f, "written by task3"
            with open(task.out3.localFileName, "w") as f:
                print >>f, "written by task3"

        wf = PypeThreadWorkflow()
        wf.addTasks([task1, task2, task3])

        wf.refreshTargets()

        outputSet = set()
        outputSet.add("written by task1")
        outputSet.add("written by task2")
        outputSet.add("written by task3")

        with open(outfileObj.localFileName) as f:
            i = 0
            for l in f:
                l = l.strip()
                assert l in outputSet
                i += 1
            assert_equal(i, 3)

    def test_stateDataObjects(self):

        infileObj =\
        pypeflow.data.PypeLocalFile("file://localhost/tmp/pypetest/test_for_shared_output_in.txt")

        outfileObj =\
        pypeflow.data.PypeLocalFile("file://localhost/tmp/pypetest/test_for_shared_output_out.txt")

        out1 =\
        pypeflow.data.PypeLocalFile("file://localhost/tmp/pypetest/test_for_shared_output_out1.txt")

        out2 =\
        pypeflow.data.PypeLocalFile("file://localhost/tmp/pypetest/test_for_shared_output_out2.txt")

        out3 =\
        pypeflow.data.PypeLocalFile("file://localhost/tmp/pypetest/test_for_shared_output_out3.txt")

        s1 =\
        pypeflow.data.PypeLocalFile("state://localhost/tmp/pypetest/.state1")

        s2 =\
        pypeflow.data.PypeLocalFile("state://localhost/tmp/pypetest/.state2")

        s3 =\
        pypeflow.data.PypeLocalFile("state://localhost/tmp/pypetest/.state3")

        import os
        import time
        os.system("rm -rf /tmp/pypetest/*")
        time.sleep(2)

        with open(infileObj.localFileName,"w") as f:
            f.write("test")

        PypeThreadWorkflow = pypeflow.controller.PypeThreadWorkflow
        PypeThreadTaskBase = pypeflow.controller.PypeThreadTaskBase
        PypeTask = pypeflow.task.PypeTask
        wf = PypeThreadWorkflow()

        @PypeTask(mutableDataObjs = {"out":outfileObj},
                  outputDataObjs = {"out1":out1, "s1":s1},
                  inputDataObjs = {"in":infileObj},
                  TaskType=PypeThreadTaskBase)
        def task1(task):
            with open(task.out.localFileName, "a") as f:
                print >>f, "written by task1"
            with open(task.s1.localFileName, "w") as f:
                print >>f, "state set"
            with open(task.out1.localFileName, "w") as f:
                print >>f, "written by task1"

        @PypeTask(mutableDataObjs = {"out":outfileObj},
                  outputDataObjs = {"out2":out2, "s2":s2},
                  inputDataObjs = {"in":infileObj, "s1":s1},
                  TaskType=PypeThreadTaskBase)
        def task2(task):
            with open(task.out.localFileName, "a") as f:
                print >>f, "written by task2"
            with open(task.s2.localFileName, "w") as f:
                print >>f, "state set"
            with open(task.out2.localFileName, "w") as f:
                print >>f, "written by task2"

        @PypeTask(mutableDataObjs = {"out":outfileObj},
                  outputDataObjs = {"out3":out3, "s3":s3},
                  inputDataObjs = {"in":infileObj, "s2":s2},
                  TaskType=PypeThreadTaskBase)
        def task3(task):
            with open(task.out.localFileName, "a") as f:
                print >>f, "written by task3"
            with open(task.s3.localFileName, "w") as f:
                print >>f, "state set"
            with open(task.out3.localFileName, "w") as f:
                print >>f, "written by task3"

        wf = PypeThreadWorkflow()
        wf.addTasks([task1, task2, task3])

        wf.refreshTargets()

        with open(outfileObj.localFileName) as f:
            i = 0
            for l in f:
                i += 1
                l = l.strip()
                assert l == "written by task%d" % i
        assert i == 3

    def test_stateDataObjects2(self):

        infileObj =\
        pypeflow.data.PypeLocalFile("file://localhost/tmp/pypetest/test_for_shared_output_in.txt")

        outfileObj =\
        pypeflow.data.PypeLocalFile("file://localhost/tmp/pypetest/test_for_shared_output_out.txt")

        out1 =\
        pypeflow.data.PypeLocalFile("file://localhost/tmp/pypetest/test_for_shared_output_out1.txt")

        out2 =\
        pypeflow.data.PypeLocalFile("file://localhost/tmp/pypetest/test_for_shared_output_out2.txt")

        out3 =\
        pypeflow.data.PypeLocalFile("file://localhost/tmp/pypetest/test_for_shared_output_out3.txt")

        s1 =\
        pypeflow.data.PypeLocalFile("state://localhost/tmp/pypetest/.state1")

        s2 =\
        pypeflow.data.PypeLocalFile("state://localhost/tmp/pypetest/.state2")

        s3 =\
        pypeflow.data.PypeLocalFile("state://localhost/tmp/pypetest/.state3")

        import os
        import time
        os.system("rm -rf /tmp/pypetest/*")
        time.sleep(2)

        with open(infileObj.localFileName,"w") as f:
            f.write("test")

        PypeThreadWorkflow = pypeflow.controller.PypeThreadWorkflow
        PypeThreadTaskBase = pypeflow.controller.PypeThreadTaskBase
        PypeTask = pypeflow.task.PypeTask
        wf = PypeThreadWorkflow()

        @PypeTask( outputDataObjs = {"out1":out1, "s1":s1},
                  inputDataObjs = {"in":infileObj},
                  TaskType=PypeThreadTaskBase)
        def task1(task):
            with open(task.s1.localFileName, "w") as f:
                print >>f, "state set"
            with open(task.out1.localFileName, "w") as f:
                print >>f, "written by task1"

        @PypeTask(outputDataObjs = {"out2":out2, "s2":s2},
                  inputDataObjs = {"in":infileObj, "s1":s1},
                  TaskType=PypeThreadTaskBase)
        def task2(task):
            with open(task.s2.localFileName, "w") as f:
                print >>f, "state set"
            with open(task.out2.localFileName, "w") as f:
                print >>f, "written by task2"

        @PypeTask(outputDataObjs = {"out3":out3, "s3":s3},
                  inputDataObjs = {"in":infileObj, "s2":s2},
                  TaskType=PypeThreadTaskBase)
        def task3(task):
            with open(task.s3.localFileName, "w") as f:
                print >>f, "state set"
            with open(task.out3.localFileName, "w") as f:
                print >>f, "written by task3"

        wf = PypeThreadWorkflow()
        wf.addTasks([task1, task2, task3])

        wf.refreshTargets([s3])

        for i in range(1,4):
            with open("/tmp/pypetest/test_for_shared_output_out%d.txt" % i) as f:
                l = f.read().strip()
                assert l == "written by task%d" % i




# @author Jason Chin
#
# Copyright (C) 2010 by Jason Chin 
# Copyright (C) 2011 by Jason Chin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import sys
import os 


from pypeflow.common import * 
from pypeflow.task import PypeThreadTaskBase, PypeTaskBase
from pypeflow.task import PypeTask, PypeShellTask, PypeSGETask, PypeDistributibleTask
from pypeflow.controller import PypeWorkflow, PypeThreadWorkflow, PypeMPWorkflow
from pypeflow.data import PypeLocalFile, makePypeLocalFile, fn
import logging
import time

logger = logging.getLogger()
#logger.setLevel(logging.INFO)
logger.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
ch.setFormatter(formatter)
logger.addHandler(ch)

inputs = {"input": makePypeLocalFile("/tmp/test1_input")}
outputs = {"output": makePypeLocalFile("/tmp/test1_output")}
os.system("touch /tmp/test1_input")

@PypeTask(inputs = inputs, outputs = outputs, TaskType = PypeThreadTaskBase)
def f(self):
    i = 0
    while 1:
        time.sleep(0.1)
        if self.shutdown_event != None and self.shutdown_event.is_set():
            break
        if i > 10:
            break
        i += 1
    if self.shutdown_event == None or not self.shutdown_event.is_set():
        os.system("touch %s" % fn(self.output))

wf = PypeThreadWorkflow()
wf.addTasks([f])
wf.refreshTargets()










import networkx as nx
from falcon_kit.fc_asm_graph import AsmGraph


def main(argv=None):
  G_asm = AsmGraph("sg_edges_list", "utg_data", "ctg_paths")


  p_ctg_coor_map = {}
  for fn in ("p_ctg_tiling_path", "a_ctg_tiling_path"):
    f = open(fn)
    for row in f:
        row = row.strip().split()
        ctg_id, v, w, edge_rid, b, e  = row[:6]
        if ctg_id not in p_ctg_coor_map:
            coor = 0   # the p_ctg_tiling_path should be sorted by contig the order of the edges in the tiling path
            p_ctg_coor_map[ctg_id] = {}
            p_ctg_coor_map[ctg_id][v] = 0
            coor += abs(int(b) - int(e))
            p_ctg_coor_map[ctg_id][w] = coor
            G_asm.node_to_ctg[w]
            print ctg_id, v, 0, " ".join(list(G_asm.node_to_ctg[v]))
            print ctg_id, w, coor, " ".join(list(G_asm.node_to_ctg[w]))
            continue
        else:
            coor += abs(int(b) - int(e))
            p_ctg_coor_map[ctg_id][w] = coor
            print ctg_id, w, coor, " ".join(list(G_asm.node_to_ctg[w]))
    f.close()



from pypeflow.common import *
from pypeflow.data import PypeLocalFile, makePypeLocalFile, fn
from pypeflow.task import PypeTask, PypeThreadTaskBase, PypeTaskBase
from pypeflow.controller import PypeWorkflow, PypeThreadWorkflow
import os
import uuid
import sys



def run_script(job_data, job_type = "SGE" ):
    if job_type == "SGE":
        job_name = job_data["job_name"]
        cwd = job_data["cwd"]
        sge_option = job_data["sge_option"]
        script_fn = job_data["script_fn"]
        sge_cmd="qsub -N {job_name} {sge_option} -o {cwd}/sge_log -j y\
                 -S /bin/bash {script}".format(job_name=job_name,
                                               cwd=os.getcwd(),
                                               sge_option=sge_option,
                                               script=script_fn)

        #print sge_cmd
        os.system( sge_cmd )
        os.system( "sleep 1")
    elif job_type == "local":
        os.system( "bash %s" % job_data["script_fn"] )

def wait_for_file(filename, task = None, job_name = ""):
    while 1:
        time.sleep(30)
        if os.path.exists(filename):
            break

        if task != None:
            if task.shutdown_event != None and task.shutdown_event.is_set():
                os.system("qdel %s" % job_name)
                break

def run_p_task(self):
    p_script_fn = self.parameters["p_file"]
    job_id = self.parameters["job_id"]
    cwd = self.parameters["cwd"]
    script_dir = os.path.join( cwd )
    script_fn =  os.path.join( script_dir , "rp_%05d.sh" % (job_id))
    log_path = os.path.join( script_dir, "rp_%05d.log" % (job_id))
    script = []
    script.append( "export PATH=~/task2014/dazzler/DALIGNER/:$PATH" )
    script.append( "cd %s" % cwd )
    script.append( ("/usr/bin/time bash %s " % p_script_fn)  + ( " >& %s " % log_path ) + ( " && touch %s" % fn( self.job_done ) ) )

    with open(script_fn,"w") as script_file:
        script_file.write("\n".join(script))



    job_name = self.URL.split("/")[-1]
    job_name += "-"+str(uuid.uuid1())[:8]
    job_data = {"job_name": job_name,
                "cwd": cwd,
                "sge_option": " -pe smp 2 -q huasm ",
                "script_fn": script_fn }
    run_script(job_data, job_type = "SGE")
    wait_for_file( fn( self.job_done ), task=self, job_name=job_name )

def run_consensus_task(self):
    job_id = self.parameters["job_id"]
    cwd = self.parameters["cwd"]
    script_dir = os.path.join( cwd )
    script_fn =  os.path.join( script_dir , "cp_%05d.sh" % (job_id))
    log_path = os.path.join( script_dir, "cp_%05d.log" % (job_id))

    with open( os.path.join(cwd, "c_%05d.sh" % job_id), "w") as p_script:
        print >> p_script, ". /mnt/secondary/Share/HBAR_03202013/bin/activate"
        print >> p_script, "cd .."
        print >> p_script, """./LA4Falcon -o -f:%s las_files/%s.%d.las | """ % (prefix, prefix, job_id),
        print >> p_script, """ falcon_sense.py --trim --output_multi --min_idt 0.70 --min_cov 4 --local_match_count_threshold 3 --max_n_read 800 --n_core 8 > %s""" % fn(self.out_file)

    script = []
    script.append( "cd %s" % cwd )
    script.append( ("/usr/bin/time bash c_%05d.sh " % job_id )  + ( " >& %s " % log_path ) + ( " && touch c_%05d_done" % job_id  ) )

    with open(script_fn,"w") as script_file:
        script_file.write("\n".join(script))

    job_name = self.URL.split("/")[-1]
    job_name += "-"+str(uuid.uuid1())[:8]
    job_data = {"job_name": job_name,
                "cwd": cwd,
                "sge_option": " -pe smp 6 -q huasm ",
                "script_fn": script_fn }
    run_script(job_data, job_type = "SGE")
    wait_for_file( os.path.join(cwd,"c_%05d_done" % job_id) , task=self, job_name=job_name )


if __name__ == "__main__":

    prefix = sys.argv[1]

    concurrent_jobs = 16
    PypeThreadWorkflow.setNumThreadAllowed(concurrent_jobs, concurrent_jobs)
    wf = PypeThreadWorkflow()

    mjob_data = {}

    with open("run_jobs.sh") as f:
        for l in f:
            l = l.strip().split()
            if l[0] not in ( "LAsort", "LAmerge" ):
                continue
            if l[0] == "LAsort":
                p_id = int( l[2].split(".")[1] )
                mjob_data.setdefault( p_id, [] )
                mjob_data[p_id].append(  " ".join(l) )
            if l[0] == "LAmerge":
                l2 = l[2].split(".")
                if l2[1] == "L2":
                    p_id = int(  l[2].split(".")[2] )
                    mjob_data.setdefault( p_id, [] )
                    mjob_data[p_id].append(  " ".join(l) )
                else:
                    p_id = int( l[2].split(".")[1] )
                    mjob_data.setdefault( p_id, [] )
                    mjob_data[p_id].append(  " ".join(l) )

    db_file = makePypeLocalFile(os.path.abspath( "./%s.db" % prefix ))

    for p_id in mjob_data:
        s_data = mjob_data[p_id]

        try:
            os.makedirs("./p_%05d" % p_id)
            os.makedirs("./p_%05d/sge_log" % p_id)
        except OSError:
            pass
        try:
            os.makedirs("./preads")
        except OSError:
            pass
        try:
            os.makedirs("./las_files")
        except OSError:
            pass
        with open("./p_%05d/p_%05d.sh" % (p_id, p_id), "w") as p_script:
            print >> p_script, """for f in `find .. -wholename "*job*/%s.%d.%s.*.*.las"`; do ln -sf $f .; done""" % (prefix, p_id, prefix)
            for l in s_data:
                print >> p_script, l
                print >> p_script, "mv %s.%d.las ../las_files" % (prefix, p_id)

        p_file = os.path.abspath( "./p_%05d/p_%05d.sh" % (p_id, p_id) )
        job_done = makePypeLocalFile(os.path.abspath( "./p_%05d/p_%05d_done" % (p_id,p_id)  ))
        parameters =  {"p_file": p_file,
                       "cwd": os.path.join(os.getcwd(), "p_%05d" % p_id),
                       "job_id": p_id}
        make_p_task = PypeTask( inputs = {"db_file": db_file},
                                       outputs = {"job_done": job_done},
                                       parameters = parameters,
                                       TaskType = PypeThreadTaskBase,
                                       URL = "task://localhost/ptask_%05d" % p_id )
        p_task = make_p_task ( run_p_task )

        wf.addTask(p_task)


        out_file = makePypeLocalFile(os.path.abspath( "./preads/out.%04d.fa" % p_id  ))
        parameters =  {"cwd": os.path.join(os.getcwd(), "preads" ),
                       "job_id": p_id}
        make_c_task = PypeTask( inputs = {"job_done": job_done},
                                outputs = {"out_file": out_file },
                                parameters = parameters,
                                TaskType = PypeThreadTaskBase,
                                URL = "task://localhost/ct_%05d" % p_id )

        c_task = make_c_task( run_consensus_task )
        wf.addTask(c_task)
        print p_id
    wf.refreshTargets(updateFreq = 15) #all



#!/usr/bin/env python2.7
"""Blocking process-watcher.

See fs_based.py. Here, delete is a no-op, and run() starts threads, so
the main program needs to wait for threads to finish somehow.

Typical submission_string:

    qsub -S /bin/bash -sync y -V -q production -N ${JOB_ID} \\\n -o "${STDOUT_FILE}" \\\n -e "${STDERR_FILE}" \\\n -pe smp ${NPROC} \\\n "${CMD}"
"""
try:
    from shlex import quote
except ImportError:
    from pipes import quote
import collections
import contextlib
import glob
import json
import logging
import os
import pprint
import re
import signal
import string
import subprocess
import sys
import threading
import time
import traceback

log = logging.getLogger(__name__)

STATE_FN = 'state.py'
Job = collections.namedtuple('Job', ['jobid', 'cmd', 'rundir', 'options'])
MetaJob = collections.namedtuple('MetaJob', ['job', 'lang_exe'])
lang_python_exe = sys.executable
lang_bash_exe = '/bin/bash'

@contextlib.contextmanager
def cd(newdir):
    prevdir = os.getcwd()
    log.debug('CD: %r <- %r' %(newdir, prevdir))
    os.chdir(os.path.expanduser(newdir))
    try:
        yield
    finally:
        log.debug('CD: %r -> %r' %(newdir, prevdir))
        os.chdir(prevdir)

class MetaJobClass(object):
    ext = {
        lang_python_exe: '.py',
        lang_bash_exe: '.bash',
    }
    def get_wrapper(self):
        # Totally by convention, for now.
        return '%s/run-%s%s' %(self.mj.job.rundir, self.mj.job.jobid, self.ext[self.mj.lang_exe])
    def get_sentinel(self):
        return 'exit-%s' %self.mj.job.jobid # in watched dir
    def get_pid(self):
        return self.mj.pid
    def kill(self, pid, sig):
        stored_pid = self.get_pid()
        if not pid:
            pid = stored_pid
            log.info('Not passed a pid to kill. Using stored pid:%s' %pid)
        if pid and stored_pid:
            if pid != stored_pid:
                log.error('pid:%s != stored_pid:%s' %(pid, stored_pid))
        os.kill(pid, sig)
    def __init__(self, mj):
        self.mj = mj
class State(object):
    def notify_threaded(self, jobid):
        self.jobids_threaded.add(jobid)
    def notify_started(self, jobid):
        #state.top['jobids_submitted'].append(jobid)
        self.jobids_submitted.add(jobid)
        self.jobids_threaded.remove(jobid)
    def notify_exited(self, jobid, rc):
        #self.top['jobid2exit'][jobid] = rc
        self.jobid2exit[jobid] = rc
        self.jobids_submitted.remove(jobid)
    def set_job(self, jobid, mjob):
        # Is this needed? For now, we are not actually saving state, so no.
        self.top['jobs'][jobid] = mjob
    def update_jobid2status(self, jobid2status):
        for jobid in self.jobids_threaded:
            status = 'THREADED'
            jobid2status[jobid] = status
        for jobid in self.jobids_submitted:
            status = 'RUNNING'
            # but actually it might not have started yet, or it could be dead, since we have blocking qsub calls
            jobid2status[jobid] = status
        for jobid, rc in self.jobid2exit.iteritems():
            status = 'EXIT {}'.format(rc)
            jobid2status[jobid] = status
    def get_running_jobids(self):
        return list(self.jobids_submitted)
    def serialize(state):
        return pprint.pformat(state.top)
    @staticmethod
    def deserialize(directory, content):
        state = State(directory)
        state.top = eval(content)
        state.content_prev = content
        return state
    @staticmethod
    def create(directory):
        state = State(directory)
        #makedirs(state.get_directory_wrappers())
        #makedirs(state.get_directory_jobs())
        return state
    def __init__(self, directory):
        self.__directory = os.path.abspath(directory)
        self.content_prev = ''
        self.top = dict() # for serialization, when we decide we need it
        self.top['jobs'] = dict()
        #self.top['jobids_submitted'] = list()
        #self.top['jobid2exit'] = dict()
        self.jobids_threaded = set()
        self.jobids_submitted = set()
        self.jobid2exit = dict()

class SafeState(object):
    """Synchronized State proxy for accessing any
    data which might be modified in a Thread.
    """
    def notify_threaded(self, jobid):
        with self.lock:
            self.state.notify_threaded(jobid)
    def notify_started(self, jobid):
        with self.lock:
            self.state.notify_started(jobid)
    def notify_exited(self, jobid, rc):
        with self.lock:
            self.state.notify_exited(jobid, rc)
    def update_jobid2status(self, table):
        with self.lock:
            return self.state.update_jobid2status(table)
    def get_running_jobids(self):
        with self.lock:
            return self.state.get_running_jobids()
    def serialize(self):
        with self.lock:
            return self.state.serialize()
    def __getattr__(self, name):
        """For all other methods, just delegate.
        """
        return getattr(self.state, name)
    def __init__(self, state):
        self.state = state
        self.lock = threading.Lock()

def get_state(directory):
    """For now, we never write.
    """
    state_fn = os.path.join(directory, STATE_FN)
    if not os.path.exists(state_fn):
        return State.create(directory)
    assert False, 'No state directory needed, for now.'
    try:
        return State.deserialize(directory, open(state_fn).read())
    except Exception:
        log.exception('Failed to read state "%s". Ignoring (and soon over-writing) current state.'%state_fn)
        # TODO: Backup previous STATE_FN?
        return State(directory)
def State_save(state):
    # TODO: RW Locks, maybe for runtime of whole program.
    content = state.serialize()
    content_prev = state.content_prev
    if content == content_prev:
        return
    fn = state.get_state_fn()
    open(fn, 'w').write(content)
    log.debug('saved state to %s' %repr(os.path.abspath(fn)))
def Job_get_MetaJob(job, lang_exe=lang_bash_exe):
    return MetaJob(job, lang_exe=lang_exe)
def MetaJob_wrap(mjob, state):
    """Write wrapper contents to mjob.wrapper.
    """
    metajob_rundir = mjob.job.rundir
    wdir = metajob_rundir

    bash_template = """#!%(lang_exe)s
cmd="%(cmd)s"
rundir="%(rundir)s"
finish() {
  echo "finish code: $?"
}
trap finish 0
#printenv
echo
set -ex
while [ ! -d "$rundir" ]; do sleep 1; done
cd "$rundir"
eval "$cmd"
    """
    mji = MetaJobClass(mjob)
    wrapper_fn = os.path.join(wdir, mji.get_wrapper())
    command = mjob.job.cmd

    wrapped = bash_template %dict(
        lang_exe=mjob.lang_exe,
        cmd=command,
        rundir=metajob_rundir,
    )
    log.debug('Writing wrapper "%s"' %wrapper_fn)
    open(wrapper_fn, 'w').write(wrapped)
    system('chmod +x {}'.format(wrapper_fn))

class JobThread(threading.Thread):
    def run(self):
        """Propagate environment, plus env_extra.
        """
        self.notify_start(self.jobname)
        log.debug('hello! started Thread {}'.format(threading.current_thread()))
        myenv = dict(os.environ)
        myenv.update(self.env_extra)
        log.debug('myenv:\n{}'.format(pprint.pformat(myenv)))
        log.info("Popen: '{}'".format(self.cmd))
        p = subprocess.Popen(self.cmd, env=myenv, shell=True)
        log.debug("pid: {}".format(p.pid))
        p.wait()
        rc = p.returncode
        log.debug("rc: {}".format(rc))
        self.notify_exit(self.jobname, rc)
    def __init__(self, jobname, cmd, notify_start, notify_exit, env_extra):
        super(JobThread, self).__init__()
        self.jobname = jobname
        self.cmd = cmd
        self.notify_start = notify_start
        self.notify_exit = notify_exit
        self.env_extra = env_extra

class StringJobSubmitter(object):
    """Substitute some variables into self.submission_string.
    Use mains/job_start.sh as the top script. That requires
    PYPEFLOW_JOB_START_SCRIPT in the environment as the real
    script to run. This way, we are guaranteed that the top script exists,
    and we can wait for the rest to appear in the filesystem.
    """
    def submit(self, jobname, mjob, state):
        """Prepare job (based on wrappers) and submit as a new thread.
        """
        state.set_job(jobname, mjob)
        jobname = mjob.job.jobid
        mji = MetaJobClass(mjob)
        #script_fn = os.path.join(state.get_directory_wrappers(), mji.get_wrapper())
        script_fn = mji.get_wrapper()
        exe = mjob.lang_exe

        state.notify_threaded(jobname)
        self.start(jobname, state, exe, script_fn) # Can raise
    def get_cmd(self, jobname, script_fn, nproc):
        """Vars:
        JOB_ID, STDOUT_FILE, STDERR_FILE, NPROC, CMD
        """
        # We wrap in a program that waits for the executable to exist, so
        # the filesystem has time to catch up on the remote machine.
        # Hopefully, this will allow dependencies to become ready as well.
        job_start_fn = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'mains/job_start.sh')
        CMD = job_start_fn
        mapping = dict()
        mapping['CMD'] = CMD
        mapping['JOB_ID'] = jobname
        mapping['STDOUT_FILE'] = script_fn + '.stdout'
        mapping['STDERR_FILE'] = script_fn + '.stderr'
        mapping['NPROC'] = str(nproc)
        log.debug('mapping, submission_string: {}, {}'.format(repr(mapping), self.submission_string))
        t = string.Template(self.submission_string)
        return t.substitute(mapping)
    def start(self, jobname, state, exe, script_fn):
        """Run job in thread.
        Thread will notify state.
        Can raise.
        """
        nproc = 4
        #cmd = script_fn
        cmd = self.get_cmd(jobname, script_fn, nproc)
        # job_start.sh relies on PYPEFLOW_*
        env_extra = {
            "PYPEFLOW_JOB_START_SCRIPT": script_fn,
            "PYPEFLOW_JOB_START_TIMEOUT": "60",
        }
        log.debug('env_extra={}'.format(pprint.pformat(env_extra)))
        notify_start = state.notify_started
        notify_exit = state.notify_exited
        th = JobThread(jobname, cmd, notify_start, notify_exit, env_extra)
        #th.setDaemon(True)
        th.start()
    def __repr__(self):
        return 'StringJobSubmitter(%s)' %repr(self.submission_string)
    def __init__(self, submission_string):
        self.submission_string = submission_string

def link_rundir(state_rundir, user_rundir):
    if user_rundir:
        link_fn = os.path.join(user_rundir, 'pwatcher.dir')
        if os.path.lexists(link_fn):
            os.unlink(link_fn)
        os.symlink(os.path.abspath(state_rundir), link_fn)

def cmd_run(state, jobids, job_type, job_queue):
    """
    Wrap them and run them locally, each in the foreground of a thread.
    """
    jobs = dict()
    submitted = list()
    result = {'submitted': submitted}
    if job_type != 'string':
        log.warning("In blocking pwatcher, job_type={!r}, should be 'string'".format(job_type))
    for jobid, desc in jobids.iteritems():
        assert 'cmd' in desc
        cmd = desc['cmd']
        if 'rundir' in desc:
            rundir = desc['rundir']
        else:
            rundir = os.path.dirname(cmd)
        options = {}
        jobs[jobid] = Job(jobid, cmd, rundir, options)
    log.debug('jobs:\n%s' %pprint.pformat(jobs))
    submission_string = job_queue
    submitter = StringJobSubmitter(submission_string)
    log.debug('submitter: {!r}'.format(submitter))
    for jobid, job in jobs.iteritems():
        desc = jobids[jobid]
        log.debug('starting job %s' %pprint.pformat(job))
        mjob = Job_get_MetaJob(job)
        MetaJob_wrap(mjob, state)
        try:
            #link_rundir(state.get_directory_job(jobid), desc.get('rundir'))
            submitter.submit(jobid, mjob, state)
            submitted.append(jobid)
        except Exception:
            log.exception('Failed to submit background-job:\n{!r}'.format(
                submitter))
    return result
    # The caller is responsible for deciding what to do about job-submission failures. Re-try, maybe?

def system(call, checked=False):
    log.info('!{}'.format(call))
    rc = os.system(call)
    if checked and rc:
        raise Exception('{} <- {!r}'.format(rc, call))
    return rc

_warned = dict()
def warnonce(hashkey, msg):
    if hashkey in _warned:
        return
    log.warning(msg)
    _warned[hashkey] = True

def cmd_query(state, which, jobids):
    """Return the state of named jobids.
    If which=='list', then query jobs listed as jobids.
    If which=='known', then query all known jobs.
    If which=='infer', same as 'known' now.
    """
    result = dict()
    jobstats = dict()
    result['jobids'] = jobstats
    if which == 'list':
        for jobid in jobids:
            jobstats[jobid] = 'UNKNOWN'
    state.update_jobid2status(jobstats)
    jobids = set(jobids)
    if which == 'list':
        for jobid in list(jobstats.keys()):
            # TODO: This might remove thousands. We should pass jobids along to update_jobid2status().
            if jobid not in jobids:
                del jobstats[jobid]
    return result
def cmd_delete(state, which, jobids):
    """Kill designated jobs, including (hopefully) their
    entire process groups.
    If which=='list', then kill all jobs listed as jobids.
    If which=='known', then kill all known jobs.
    If which=='infer', then kill all jobs with heartbeats.
    """
    log.error('Noop. We cannot kill blocked threads. Hopefully, everything will die on SIGTERM.')
def makedirs(path):
    if not os.path.isdir(path):
        os.makedirs(path)
def readjson(ifs):
    """Del keys that start with ~.
    That lets us have trailing commas on all other lines.
    """
    content = ifs.read()
    log.debug('content:%s' %repr(content))
    jsonval = json.loads(content)
    #pprint.pprint(jsonval)
    def striptildes(subd):
        if not isinstance(subd, dict):
            return
        for k,v in subd.items():
            if k.startswith('~'):
                del subd[k]
            else:
                striptildes(v)
    striptildes(jsonval)
    #pprint.pprint(jsonval)
    return jsonval

class ProcessWatcher(object):
    def run(self, jobids, job_type, job_queue):
        #import traceback; log.debug(''.join(traceback.format_stack()))
        log.debug('run(jobids={}, job_type={}, job_queue={})'.format(
            '<%s>'%len(jobids), job_type, job_queue))
        return cmd_run(self.state, jobids, job_type, job_queue)
    def query(self, which='list', jobids=[]):
        log.debug('query(which={!r}, jobids={})'.format(
            which, '<%s>'%len(jobids)))
        return cmd_query(self.state, which, jobids)
    def delete(self, which='list', jobids=[]):
        log.debug('delete(which={!r}, jobids={})'.format(
            which, '<%s>'%len(jobids)))
        return cmd_delete(self.state, which, jobids)
    def __init__(self, state):
        # state must be thread-safe
        self.state = state

def get_process_watcher(directory):
    state = get_state(directory)
    state = SafeState(state) # thread-safe proxy
    #log.debug('state =\n%s' %pprint.pformat(state.top))
    return ProcessWatcher(state)
    #State_save(state)

@contextlib.contextmanager
def process_watcher(directory):
    """This will (someday) hold a lock, so that
    the State can be written safely at the end.
    """
    state = get_state(directory)
    state = SafeState(state) # thread-safe proxy
    #log.debug('state =\n%s' %pprint.pformat(state.top))
    yield ProcessWatcher(state)
    #State_save(state)

def main(prog, cmd, state_dir='mainpwatcher', argsfile=None):
    logging.basicConfig()
    logging.getLogger().setLevel(logging.NOTSET)
    log.warning('logging basically configured')
    log.debug('debug mode on')
    assert cmd in ['run', 'query', 'delete']
    ifs = sys.stdin if not argsfile else open(argsfile)
    argsdict = readjson(ifs)
    log.info('argsdict =\n%s' %pprint.pformat(argsdict))
    with process_watcher(state_dir) as watcher:
        result = getattr(watcher, cmd)(**argsdict)
        if result is not None:
            log.info('getattr({!r}, {!r}): {}'.format(
                watcher, cmd, pprint.pformat(result)))
        log.info('Waiting for running jobs...r')
        while watcher.state.get_running_jobids():
            log.info('running: {!s}'.format(watcher.state.get_running_jobids()))
            time.sleep(1)

if __name__ == "__main__":
    #import pdb
    #pdb.set_trace()
    main(*sys.argv)



#!/usr/bin/env python2.7
from .. import stats_preassembly
import argparse
import json
import logging

log = logging.getLogger()

def do_report(db, preads_fofn, genome_length, length_cutoff, out):
    kwds = dict(
        i_preads_fofn_fn=preads_fofn,
        i_raw_reads_db_fn=db,
        genome_length=genome_length,
        length_cutoff=length_cutoff,
    )
    report_dict = stats_preassembly.calc_dict(**kwds)
    content = json.dumps(report_dict, sort_keys=True, indent=4, separators=(',', ': '))
    open(out, 'w').write(content)

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--genome-length',
        type=int,
        required=True,
        help='Estimated number of bases in the full genome haplotype.')
    parser.add_argument('--length-cutoff',
        type=int,
        required=True,
        help='Minimum length of any seed read.')
    parser.add_argument('--db',
        required=True,
        help='Path to raw_reads.db (dazzler DB)')
    parser.add_argument('--preads-fofn',
        required=True,
        help='Path to FOFN of preads fasta files.')
    parser.add_argument('--out',
        required=True,
        help='Path to JSON output file.')
    ARGS = parser.parse_args()
    do_report(**vars(ARGS))


if __name__ == "__main__":
    logging.basicConfig()
    log.setLevel(logging.DEBUG)
    main()



from .simple_pwatcher_bridge import (
        PypeProcWatcherWorkflow, MyFakePypeThreadTaskBase,
        makePypeLocalFile, fn, PypeTask)
PypeThreadTaskBase = MyFakePypeThreadTaskBase

__all__ = [
        'PypeProcWatcherWorkflow', 'PypeThreadTaskBase',
        'makePypeLocalFile', 'fn', 'PypeTask',
]



from pypeflow.common import *
from pypeflow.data import PypeLocalFile, makePypeLocalFile, fn
from pypeflow.task import PypeTask, PypeThreadTaskBase, PypeTaskBase
from pypeflow.controller import PypeWorkflow, PypeThreadWorkflow
import os
import uuid
import sys


def run_script(job_data, job_type = "SGE" ):
    if job_type == "SGE":
        job_name = job_data["job_name"]
        cwd = job_data["cwd"]
        sge_option = job_data["sge_option"]
        script_fn = job_data["script_fn"]
        sge_cmd="qsub -N {job_name} {sge_option} -o {cwd}/sge_log -j y\
                 -S /bin/bash {script}".format(job_name=job_name,
                                               cwd=os.getcwd(),
                                               sge_option=sge_option,
                                               script=script_fn)

        #print sge_cmd
        os.system( sge_cmd )
        os.system( "sleep 1")
    elif job_type == "local":
        os.system( "bash %s" % job_data["script_fn"] )

def wait_for_file(filename, task = None, job_name = ""):
    while 1:
        time.sleep(60)
        if os.path.exists(filename):
            break

        if task != None:
            if task.shutdown_event != None and task.shutdown_event.is_set():
                os.system("qdel %s" % job_name)
                break

def run_daligner(self):
    daligner_cmd = self.parameters["daligner_cmd"]
    job_id = self.parameters["job_id"]
    cwd = self.parameters["cwd"]
    script_dir = os.path.join( cwd )
    script_fn =  os.path.join( script_dir , "rj_%05d.sh" % (job_id))
    log_path = os.path.join( script_dir, "rj_%05d.log" % (job_id))
    script = []
    script.append( "export PATH=~/task2014/dazzler/DALIGNER/:$PATH" )
    script.append( "cd %s" % cwd )
    script.append( "/usr/bin/time "+ daligner_cmd + ( " >& %s " % log_path ) + ( " && touch %s" % fn( self.job_done ) ) )

    with open(script_fn,"w") as script_file:
        script_file.write("\n".join(script))

    job_name = self.URL.split("/")[-1]
    job_name += "-"+str(uuid.uuid1())[:8]
    job_data = {"job_name": job_name,
                "cwd": cwd,
                "sge_option": " -pe smp 6 -q huasm ",
                "script_fn": script_fn }
    run_script(job_data, job_type = "SGE")
    wait_for_file( fn( self.job_done ), task=self, job_name=job_name )

if __name__ == "__main__":
    prefix = sys.argv[1]
    concurrent_jobs = 64
    PypeThreadWorkflow.setNumThreadAllowed(concurrent_jobs, concurrent_jobs)
    wf = PypeThreadWorkflow()

    job_id = 0
    db_file = makePypeLocalFile(os.path.abspath( "./%s.db" % prefix ))
    with open("run_jobs.sh") as f :
        for l in f :
            l = l.strip().split()
            if l[0] == "daligner":
                try:
                    os.makedirs("./job_%05d" % job_id)
                except OSError:
                    pass
                os.system("cd ./job_%05d;ln -s ../.%s.bps .; ln -s ../.%s.idx .; ln -s ../%s.db ." % (job_id, prefix, prefix, prefix) )
                job_done = makePypeLocalFile(os.path.abspath( "./job_%05d/job_%05d_done" % (job_id,job_id)  ))
                parameters =  {"daligner_cmd": " ".join(l),
                               "cwd": os.path.join(os.getcwd(), "job_%05d" % job_id),
                               "job_id": job_id}
                make_daligner_task = PypeTask( inputs = {"db_file": db_file},
                                               outputs = {"job_done": job_done},
                                               parameters = parameters,
                                               TaskType = PypeThreadTaskBase,
                                               URL = "task://localhost/mtask_%05d" % job_id )
                daligner_task = make_daligner_task ( run_daligner )
                wf.addTask(daligner_task)
                job_id += 1
                print job_id
    wf.refreshTargets(updateFreq = 45) #all



#!/usr/bin/env python
"""Usage:

    LAsort.py DB <args>

Run LAcheck on each input in args. Exclude any failures from
the arglist. Then run LAsort on the remaining arglist.
"""
import sys, os

def log(msg):
    sys.stderr.write(msg + '\n')

def system(call, checked=False):
    log('!{}'.format(call))
    rc = os.system(call)
    if rc:
        msg = '{} <- {!r}'.format(rc, call)
        if checked:
            raise Exception(msg)
        log(msg)
    return rc

def main(argv=sys.argv):
    log('argv:{!r}'.format(argv))
    db = argv[1]
    args = argv[2:] # Skip program name
    lass = list()
    new_args = list()
    new_args.append('LAsort')
    for arg in args:
        if arg.startswith('-'):
            new_args.append(arg)
        else:
            lass.append(arg)
    for las in lass:
        rc = system('LAcheck -v {} {}.las'.format(db, las))
        if rc:
            log('Skipping {}.las'.format(las))
        else:
            new_args.append(las)
    system(' '.join(new_args))


if __name__ == "__main__":
    main()



#!/usr/bin/env python

#################################################################################$$
# Copyright (c) 2011-2014, Pacific Biosciences of California, Inc.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted (subject to the limitations in the
# disclaimer below) provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright
#  notice, this list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above
#  copyright notice, this list of conditions and the following
#  disclaimer in the documentation and/or other materials provided
#  with the distribution.
#
#  * Neither the name of Pacific Biosciences nor the names of its
#  contributors may be used to endorse or promote products derived
#  from this software without specific prior written permission.
#
# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE
# GRANTED BY THIS LICENSE. THIS SOFTWARE IS PROVIDED BY PACIFIC
# BIOSCIENCES AND ITS CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL PACIFIC BIOSCIENCES OR ITS
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#################################################################################$$

import networkx as nx
from pbcore.io import FastaReader

def neighbor_bound(G, v, w, radius):
    g1 = nx.ego_graph(G, v, radius=radius, undirected=False)
    g2 = nx.ego_graph(G, w, radius=radius, undirected=False)
    if len(g1) < radius or len(g2) < radius:
        return True
    print v, len(g1), w, len(g2), radius
    if len(set(g1.edges()) & set(g2.edges())) > 0:
        return True
    else:
        return False

def is_branch_node(G, n):
    out_edges = G.out_edges([n])
    n2 = [ e[1] for e in out_edges ]
    is_branch = False
    for i in range(len(n2)):
        for j in range(i+1, len(n2)):
            v = n2[i]
            w = n2[j]
            if neighbor_bound(G, v, w, 20) == False:
                is_branch = True
                break
        if is_branch == True:
            break
    return is_branch


def get_r_path(r_edges, u_path):
    tiling_path = []
    pos = 0

    for i in range( len(u_path) - 1):
        v, w = u_path[i:i+2]
        r_edge_label, overlap = r_edges[ (v, w) ]
        r_edge_seq_id, range_ = r_edge_label.split(":")
        range_ = range_.split("-")
        s, e = int(range_[0]), int(range_[1])
        pos += abs(e-s)
        tiling_path.append( (pos, w, s, e) )
    return tiling_path

def get_seq(u_edges, r_edges, path):
    subseqs = []
    pos = []
    cur_pos = 0
    full_tiling_path = []

    for i in range( len(path) - 1):
        v, w = path[i:i+2]
        pos.append( (v, cur_pos) )
        uedges = u_edges[ (v, w) ]
        uedges.sort( key= lambda x: len(x[0]) )
        subseqs.append( uedges[-1][1] )
        r_path = get_r_path( r_edges, uedges[-1][0].split("-") )
        r_path = [ ( x[0] + cur_pos, x[1], x[2], x[3]) for x in r_path ]
        full_tiling_path.extend( r_path )
        cur_pos += len( uedges[-1][1] )
    pos.append( (w, cur_pos) )
    return "".join(subseqs), pos, full_tiling_path


u_edges = {}
with open("unit_edges.dat") as f:
    for l in f:
        v, w, path, seq = l.strip().split()
        u_edges.setdefault( (v, w), [] )
        u_edges[ (v, w) ].append( (path, seq) )
len(u_edges)


r_edges = {}
with open("edges_list") as f:
    for l in f:
        v, w, edge_label, overlap = l.strip().split()
        r_edges[ (v, w) ] = (edge_label, int(overlap) )


primary_tigs_path = {}
primary_path_graph = nx.DiGraph()
begin_nodes = {}
end_nodes ={}
with open("primary_tigs_paths") as f:
    for l in f:
        l = l.strip().split()
        name = l[0][1:]
        path = l[1:]
        primary_tigs_path[name] = path
        if len(path) < 3:
            continue
        for i in range(len(path)-1):
            n1 = path[i].split(":")[0]
            n2 = path[i+1].split(":")[0]
            primary_path_graph.add_edge( n1, n2)
        begin_nodes.setdefault(path[0], [])
        begin_nodes[path[0]].append( name )
        end_nodes.setdefault(path[-1], [])
        end_nodes[path[-1]].append( name )



path_names = primary_tigs_path.keys()
path_names.sort()
primary_path_graph_r = primary_path_graph.reverse()
path_f = open("primary_tigs_paths_c","w")
pos_f = open("primary_tigs_node_pos_c", "w")
tiling_path_f = open("all_tiling_path_c", "w")
with open("primary_tigs_c.fa","w") as out_f:
    for name in path_names:
        sub_idx = 0
        c_path = [ primary_tigs_path[name][0] ]
        for v in primary_tigs_path[name][1:]:
            break_path = False

            vn = v.split(":")[0]

            if primary_path_graph.out_degree(vn) > 1:
                break_path = is_branch_node(primary_path_graph, vn)
            if primary_path_graph.in_degree(vn) > 1:
                break_path = is_branch_node(primary_path_graph_r, vn)
            if break_path:
                c_path.append(v)
                seq, pos, full_tiling_path = get_seq(u_edges, r_edges, c_path)
                for p, w, s, e in full_tiling_path:
                    print >> tiling_path_f, "%s_%02d" % (name, sub_idx), p, w, s, e
                #if len(full_tiling_path) <= 5:
                #    continue
                print >>out_f, ">%s_%02d" % (name, sub_idx)
                print >>out_f, seq
                print >>path_f, ">%s_%02d" % (name, sub_idx), " ".join(c_path)
                #print c_path
                for node, p in pos:
                    print >> pos_f, "%s_%02d %s %d" % (name, sub_idx, node, p)
                c_path = [v]
                sub_idx += 1
            else:
                c_path.append(v)

        if len(c_path) > 1:
            seq, pos, full_tiling_path = get_seq(u_edges, r_edges, c_path)
            for p, w, s, e in full_tiling_path:
                print >> tiling_path_f, "%s_%02d" % (name, sub_idx), p, w, s, e
            if len(full_tiling_path) <= 5:
                continue
            print >>out_f, ">%s_%02d" % (name, sub_idx)
            print >>out_f, seq
            print >>path_f, ">%s_%02d" % (name, sub_idx), " ".join(c_path)
            for node, p in pos:
                print >> pos_f, "%s_%02d %s %d" % (name, sub_idx, node, p)

with open("all_tigs_paths") as f:
    for l in f:
        l = l.strip().split()
        name = l[0][1:]
        name = name.split("-")
        if name[1] == "0000":
            continue
        if len(name) == 2:
            path = l[1:]
            seq, pos, full_tiling_path = get_seq(u_edges, r_edges, path)
            for p, w, s, e in full_tiling_path:
                print >> tiling_path_f, "%s" % ("-".join(name)), p, w, s, e
        else:
            path = l[1:]
            full_tiling_path = get_r_path(r_edges, path)
            for p, w, s, e in full_tiling_path:
                print >> tiling_path_f, "%s" % ("-".join(name)), p, w, s, e


path_f.close()
tiling_path_f.close()
pos_f.close()



from __future__ import absolute_import
from pypeflow.simple_pwatcher_bridge import (PypeProcWatcherWorkflow, MyFakePypeThreadTaskBase,
        makePypeLocalFile, fn, PypeTask)
PypeThreadTaskBase = MyFakePypeThreadTaskBase
from .. import pype_tasks
import argparse
import glob
import logging
import sys
import subprocess as sp
import shlex
import os

LOG = logging.getLogger(__name__)

def make_dirs(d):
    if not os.path.isdir(d):
        LOG.debug('mkdirs {}'.format(d))
        os.makedirs(d)

def get_read_ctg_map(rawread_dir, pread_dir, asm_dir):
    read_map_dir = os.path.abspath(os.path.join(asm_dir, 'read_maps'))
    make_dirs(read_map_dir)

    wf = PypeProcWatcherWorkflow(
            max_jobs=12,
    )
    """
            job_type=config['job_type'],
            job_queue=config['job_queue'],
            sge_option=config.get('sge_option', ''),
            watcher_type=config['pwatcher_type'],
            watcher_directory=config['pwatcher_directory'])
    """

    rawread_db = makePypeLocalFile(os.path.join(rawread_dir, 'raw_reads.db'))
    rawread_id_file = makePypeLocalFile(os.path.join(read_map_dir, 'dump_rawread_ids', 'rawread_ids'))

    task = PypeTask(
              inputs = {'rawread_db': rawread_db},
              outputs = {'rawread_id_file': rawread_id_file},
    )
    wf.addTask(task(pype_tasks.task_dump_rawread_ids))

    pread_db = makePypeLocalFile(os.path.join(pread_dir, 'preads.db'))
    pread_id_file = makePypeLocalFile(os.path.join(read_map_dir, 'dump_pread_ids', 'pread_ids'))

    task = PypeTask(
               inputs = {'pread_db': pread_db},
               outputs = {'pread_id_file': pread_id_file},
    )
    wf.addTask(task(pype_tasks.task_dump_pread_ids))

    wf.refreshTargets() # block

    sg_edges_list = makePypeLocalFile( os.path.join(asm_dir, 'sg_edges_list') )
    utg_data = makePypeLocalFile( os.path.join(asm_dir, 'utg_data') )
    ctg_paths = makePypeLocalFile( os.path.join(asm_dir, 'ctg_paths') )

    inputs = { 'rawread_id_file': rawread_id_file,
               'pread_id_file': pread_id_file,
               'sg_edges_list': sg_edges_list,
               'utg_data': utg_data,
               'ctg_paths': ctg_paths }

    read_to_contig_map = makePypeLocalFile(os.path.join(read_map_dir, 'get_ctg_read_map', 'read_to_contig_map'))

    task = PypeTask(
              inputs = inputs,
              outputs = {'read_to_contig_map': read_to_contig_map},
    )
    wf.addTask(task(pype_tasks.task_generate_read_to_ctg_map))

    wf.refreshTargets() # block

def parse_args(argv):
    parser = argparse.ArgumentParser(description='generate `2-asm-falcon/read_maps/read_to_contig_map` that contains the \
information from the chain of mapping: (contig id) -> (internal p-read id) -> (internal raw-read id) -> (original read id)',
              formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--basedir', type=str, default='./', help='the base working dir of a FALCON assembly')
    args = parser.parse_args(argv[1:])
    return args

def main(argv=sys.argv):
    logging.basicConfig()
    args = parse_args(argv)
    basedir = args.basedir
    rawread_dir = os.path.abspath(os.path.join( basedir, '0-rawreads'))
    pread_dir = os.path.abspath(os.path.join( basedir, '1-preads_ovl'))
    asm_dir = os.path.abspath(os.path.join( basedir, '2-asm-falcon'))

    get_read_ctg_map(rawread_dir=rawread_dir, pread_dir=pread_dir, asm_dir=asm_dir)

if __name__ == '__main__':
    main()



#!/usr/bin/env python2.7
from . import do_support, util
import argparse
import contextlib
import importlib
import inspect
import json
import logging
import os
import pprint
import sys
import time
DONE = 'done'
STATUS = 'status'
TIMEOUT = 30
LOG = logging.getLogger()
DESCRIPTION = """Given a JSON description, call a python-function.
"""
EPILOG = """
The JSON looks like this:
{
    "inputs": {"input-name": "filename"},
    "outputs": {"output-name": "output-filename (relative)"},
    "python_function": "falcon_kit.mains.foo",
    "parameters": {}
}

This program will run on the work host, and it will do several things:
    - Run in CWD.
    - Verify that inputs are available. (Wait til timeout if not.)
    - Possibly, cd to tmpdir and create symlinks from inputs.
    - Run the python-function.
      - Its module must be available (e.g. in PYTHONPATH).
      - Pass a kwd-dict of the union of inputs/outputs/parameters.
      - Ignore return-value. Expect exceptions.
    - Possibly, mv outputs from tmpdir to workdir.
    - Write exit-code into STATUS.
    - Touch DONE on success.
"""
"""
(Someday, we might also support runnable Python modules, or even executables via execvp().)

Note: qsub will *not* run this directly. There is a higher layer.
"""

def get_parser():
    class _Formatter(argparse.RawDescriptionHelpFormatter, argparse.ArgumentDefaultsHelpFormatter):
        pass
    parser = argparse.ArgumentParser(description=DESCRIPTION, epilog=EPILOG,
        formatter_class=_Formatter,
    )
    parser.add_argument('--timeout',
        type=int, default=60,
        help='How many seconds to wait for input files (and JSON) to exist. (default: %(default)s')
    parser.add_argument('--tmpdir',
        help='Root directory to run in. (Sub-dir name will be based on CWD.)')
    parser.add_argument('json_fn',
        help='JSON file, as per epilog.')
    return parser

@contextlib.contextmanager
def cd(newdir):
    prevdir = os.getcwd()
    LOG.debug('CD: %r <- %r' %(newdir, prevdir))
    os.chdir(os.path.expanduser(newdir))
    try:
        yield
    finally:
        LOG.debug('CD: %r -> %r' %(newdir, prevdir))
        os.chdir(prevdir)

def mkdirs(path):
    if not os.path.isdir(path):
        cmd = 'mkdir -p {}'.format(path)
        util.system(cmd)
def rmdirs(path):
    if os.path.isdir(path):
        if len(path) < 20 and 'home' in path:
            LOG.error('Refusing to rm {!r} since it might be your homedir.'.format(path))
            return
        cmd = 'rm -rf {}'.format(path)
        util.system(cmd)

def wait_for(fn):
    global TIMEOUT
    LOG.debug('Checking existence of {!r} with timeout={}'.format(fn, TIMEOUT))
    while not os.path.exists(fn):
        if TIMEOUT > 0:
            time.sleep(1)
            TIMEOUT -= 1
        else:
            raise Exception('Timed out waiting for {!r}'.format(fn))
    assert os.access(fn, os.R_OK), '{!r} not readable'.format(fn)

def get_func(python_function):
    mod_name, func_name = os.path.splitext(python_function)
    func_name = func_name[1:] # skip dot
    mod = importlib.import_module(mod_name)
    func = getattr(mod, func_name)
    return func

class OldTaskRunner(object):
    def __init__(self, inputs, outputs, parameters):
        for k,v in (inputs.items() + outputs.items()):
            setattr(self, k, v)
        self.parameters = parameters
        self.inputs = inputs
        self.outputs = outputs

def run_python_func(func, inputs, outputs, parameters):
    if False:
        kwds = dict()
        kwds.update(inputs)
        kwds.update(outputs)
        kwds.update(parameters)
        func(**kwds)
    else:
        # old way, for now
        cwd = os.getcwd()
        parameters['cwd'] = cwd
        self = OldTaskRunner(inputs, outputs, parameters)
        func(self=self)
        script_fn = getattr(self, 'generated_script_fn', None)
        if script_fn is not None:
            do_support.run_bash(script_fn)
def run(json_fn, timeout, tmpdir):
    if isinstance(timeout, int):
        global TIMEOUT
        TIMEOUT = timeout
    wait_for(json_fn)
    LOG.debug('Loading JSON from {!r}'.format(json_fn))
    cfg = json.loads(open(json_fn).read())
    LOG.debug(pprint.pformat(cfg))
    rundir = os.path.dirname(json_fn)
    with cd(rundir):
        run_cfg_in_tmpdir(cfg, tmpdir)
def run_cfg_in_tmpdir(cfg, tmpdir):
    for fn in cfg['inputs'].values():
        wait_for(fn)
    inputs = cfg['inputs']
    outputs = cfg['outputs']
    parameters = cfg['parameters']
    python_function_name = cfg['python_function']
    func = get_func(python_function_name)
    myinputs = dict(inputs)
    myoutputs = dict(outputs)
    finaloutdir = os.getcwd()
    if tmpdir:
        import getpass
        user = getpass.getuser()
        pid = os.getpid()
        myrundir = '{tmpdir}/{user}/pypetmp/{finaloutdir}'.format(**locals())
        rmdirs(myrundir)
        mkdirs(myrundir)
        # TODO(CD): Copy inputs w/ flock.
    else:
        myrundir = finaloutdir
    with util.cd(myrundir):
        try:
            run_python_func(func, myinputs, myoutputs, parameters)
        except TypeError:
            # Report the actual function spec.
            LOG.error('For function "{}", {}'.format(python_function_name, inspect.getargspec(func)))
            raise
    if tmpdir:
        """
        for k,v in outputs.iteritems():
            cmd = 'mv -f {} {}'.format(
                os.path.join(myrundir, v),
                os.path.join(finaloutdir, v))
            util.system(cmd)
        """
        cmd = 'rsync -av {}/ {}; rm -rf {}'.format(myrundir, finaloutdir, myrundir)
        util.system(cmd)
    for fn in cfg['outputs'].values():
        wait_for(fn)

def main():
    parser = get_parser()
    parsed_args = parser.parse_args(sys.argv[1:])
    try:
        run(**vars(parsed_args))
    except Exception:
        LOG.critical('Error in {} with args={!r}'.format(sys.argv[0], pprint.pformat(vars(parsed_args))))
        raise

if __name__ == "__main__":
    do_support.setup_simple_logging(**os.environ)
    LOG.debug('Running "{}"'.format(' '.join(sys.argv)))
    main()



#!/usr/bin/env python2.7

from setuptools import setup, Extension
import subprocess, sys
import glob

install_requires=[
        "networkx >=1.7, <=1.10",
        #"logging_tree",
        #"pbcore >= 0.6.3",
        ]

scripts = glob.glob("src/py_scripts/*.py")

try:
    local_version = '+git.{}'.format(subprocess.check_output('git rev-parse HEAD', shell=True))
except Exception:
    local_version = ''

setup(name='falcon_kit',
      version='0.7' + local_version,
      description='a small toolkit for DNA seqeucne alignment, overlapping, and assembly',
      author='Jason Chin',
      author_email='jchin@pacificbiosciences.com',
      packages=['falcon_kit',
          'falcon_kit.mains',
          'falcon_kit.util',
          ],
      package_dir={'falcon_kit':'falcon_kit/'},
      ext_modules=[
                   Extension('ext_falcon', ['src/c/ext_falcon.c', 'src/c/DW_banded.c', 'src/c/kmer_lookup.c', 'src/c/falcon.c'],
                    extra_link_args=[],
                    extra_compile_args=['-fPIC', '-O3', '-fno-omit-frame-pointer'],
                    # '-fno-omit-frame-pointer' can help with gperftools.
                    #libraries=['profiler'],
                    #include_dirs=['/home/cdunn/local/include'],
                    #library_dirs=['/home/cdunn/local/lib'],
                    #language="c++", # c for now
                    #export_symbols=['generate_consensus'], # for windows?
                   ),
                  ],
      entry_points = {'console_scripts': [
          'falcon-task=falcon_kit.mains.tasks:main',
          'fc_actg_coordinate=falcon_kit.mains.actg_coordinate:main',
          'fc_consensus=falcon_kit.mains.consensus:main',
          'fc_contig_annotate=falcon_kit.mains.contig_annotate:main',
          'fc_ctg_link_analysis=falcon_kit.mains.ctg_link_analysis:main',
          'fc_dedup_a_tigs=falcon_kit.mains.dedup_a_tigs:main',
          'fc_graph_to_contig=falcon_kit.mains.graph_to_contig:main',
          'fc_graph_to_utgs=falcon_kit.mains.graph_to_utgs:main',
          'fc_ovlp_filter=falcon_kit.mains.ovlp_filter:main',
          'fc_ovlp_stats=falcon_kit.mains.ovlp_stats:main',
          'fc_ovlp_to_graph=falcon_kit.mains.ovlp_to_graph:main',
          'fc_calc_cutoff=falcon_kit.mains.calc_cutoff:main',
          'fc_run=falcon_kit.mains.run1:main',
          'fc_run1=falcon_kit.mains.run1:main',
          'fc_fasta2fasta=falcon_kit.mains.fasta2fasta:main',
          'fc_fetch_reads=falcon_kit.mains.fetch_reads:main',
          'fc_get_read_ctg_map=falcon_kit.mains.get_read_ctg_map:main',
          'fc_pr_ctg_track=falcon_kit.mains.pr_ctg_track:main',
          'fc_rr_ctg_track=falcon_kit.mains.rr_ctg_track:main',
          ],
      },
      extras_require = {
          'falcon-task':  ['falcon_kit'],
      },
      scripts = scripts,
      zip_safe = False,
      setup_requires=install_requires,
      install_requires=install_requires
     )



from nose.tools import assert_equal
from nose import SkipTest
import pypeflow.task
import pypeflow.data

class TestPypeTaskBase:
    def test___call__(self):
        # pype_task_base = PypeTaskBase(URL, *argv, **kwargv)
        # assert_equal(expected, pype_task_base.__call__(*argv, **kwargv))
        raise SkipTest # TODO: implement your test here

    def test___init__(self):
        # pype_task_base = PypeTaskBase(URL, *argv, **kwargv)
        raise SkipTest # TODO: implement your test here

    def test_finalize(self):
        # pype_task_base = PypeTaskBase(URL, *argv, **kwargv)
        # assert_equal(expected, pype_task_base.finalize())
        raise SkipTest # TODO: implement your test here

    def test_setInputs(self):
        # pype_task_base = PypeTaskBase(URL, *argv, **kwargv)
        # assert_equal(expected, pype_task_base.setInputs(inputDataObjs))
        raise SkipTest # TODO: implement your test here

    def test_setOutputs(self):
        # pype_task_base = PypeTaskBase(URL, *argv, **kwargv)
        # assert_equal(expected, pype_task_base.setOutputs(outputDataObjs))
        raise SkipTest # TODO: implement your test here

    def test_setReferenceMD5(self):
        # pype_task_base = PypeTaskBase(URL, *argv, **kwargv)
        # assert_equal(expected, pype_task_base.setReferenceMD5(md5Str))
        raise SkipTest # TODO: implement your test here

    def test_status(self):
        # pype_task_base = PypeTaskBase(URL, *argv, **kwargv)
        # assert_equal(expected, pype_task_base.status())
        raise SkipTest # TODO: implement your test here

class TestPypeThreadTaskBase:
    def test___call__(self):
        # pype_thread_task_base = PypeThreadTaskBase()
        # assert_equal(expected, pype_thread_task_base.__call__(*argv, **kwargv))
        raise SkipTest # TODO: implement your test here

    def test_nSlots(self):
        # pype_thread_task_base = PypeThreadTaskBase()
        # assert_equal(expected, pype_thread_task_base.nSlots())
        raise SkipTest # TODO: implement your test here

    def test_setMessageQueue(self):
        # pype_thread_task_base = PypeThreadTaskBase()
        # assert_equal(expected, pype_thread_task_base.setMessageQueue(q))
        raise SkipTest # TODO: implement your test here

class TestPypeDistributiableTaskBase:
    def test___init__(self):
        # pype_distributiable_task_base = PypeDistributiableTaskBase(URL, *argv, **kwargv)
        raise SkipTest # TODO: implement your test here


class TestPypeTask:
    def test_pype_task(self):
        # assert_equal(expected, PypeTask(*argv, **kwargv))
        raise SkipTest # TODO: implement your test here

class TestPypeShellTask:
    def test_pype_shell_task(self):
        # assert_equal(expected, PypeShellTask(*argv, **kwargv))
        raise SkipTest # TODO: implement your test here

class TestPypeSGETask:
    def test_pype_sge_task(self):
        # assert_equal(expected, PypeSGETask(*argv, **kwargv))
        raise SkipTest # TODO: implement your test here

class TestPypeDistributibleTask:
    def test_pype_distributible_task(self):
        # assert_equal(expected, PypeDistributibleTask(*argv, **kwargv))
        raise SkipTest # TODO: implement your test here


class TestTimeStampCompare:
    def test_time_stamp_compare(self):
        # assert_equal(expected, timeStampCompare(inputDataObjs, outputDataObjs, parameters))
        raise SkipTest # TODO: implement your test here

class TestPypeTaskCollectionBase:
    def test___init__(self):
        # pype_task_collection_base = PypeTaskCollectionBase(URL, tasks)
        raise SkipTest # TODO: implement your test here

    def test_getTasks(self):
        # pype_task_collection_base = PypeTaskCollectionBase(URL, tasks)
        # assert_equal(expected, pype_task_collection_base.getTasks())
        raise SkipTest # TODO: implement your test here

class TestPypeTaskCollection:
    def test___init__(self):
        # pype_task_collection = PypeTaskCollection(URL, tasks)
        raise SkipTest # TODO: implement your test here

    def test_addTask(self):
        # pype_task_collection = PypeTaskCollection(URL, tasks)
        # assert_equal(expected, pype_task_collection.addTask(task))
        raise SkipTest # TODO: implement your test here

    def test_getTasks(self):
        # pype_task_collection = PypeTaskCollection(URL, tasks)
        # assert_equal(expected, pype_task_collection.getTasks())
        raise SkipTest # TODO: implement your test here

class TestPypeScatteredTasks:

    def test_pype_scattered_tasks(self):
        import os
        #os.system("rm -rf /tmp/pypetest/*")
        nChunk = 5 

        infileObj =\
        pypeflow.data.PypeSplittableLocalFile(
                      "splittablefile://localhost/tmp/pypetest/test_in_1.txt", 
                      nChunk = nChunk)

        with open(infileObj.localFileName, "w") as f:
            for i in range(nChunk):
                f.write("file%02d\n" % i)

        def scatter(*argv, **kwargv):
            outputObjs = sorted( kwargv["outputDataObjs"].items() )
            nOut = len(outputObjs)
            outputObjs = [ (o[0], o[1], open(o[1].localFileName, "w")) for o in outputObjs]
            with open(kwargv["inputDataObjs"]["completeFile"].localFileName,"r") as f:
                i = 0
                for l in f:
                   outf = outputObjs[i % nOut][2]
                   outf.write(l)
                   i += 1
            for o in outputObjs:
                o[2].close()

        PypeShellTask = pypeflow.task.PypeShellTask
        PypeTask = pypeflow.task.PypeTask
        PypeTaskBase = pypeflow.task.PypeTaskBase
        infileObj.setScatterTask(PypeTask, PypeTaskBase, scatter)
        infileObj.getScatterTask()()
        
        def gather(*argv, **kwargv):
            inputObjs = sorted( kwargv["inputDataObjs"].items() )
            with open(kwargv["outputDataObjs"]["completeFile"].localFileName,"w") as outf:
                for k, subfile in inputObjs:
                    f = open(subfile.localFileName)
                    outf.write(f.read())
                    f.close()

        outfileObj =\
        pypeflow.data.PypeSplittableLocalFile(
                      "splittablefile://localhost/tmp/pypetest/test_out_1.txt", 
                      nChunk = nChunk)

        outfileObj.setGatherTask(PypeTask, PypeTaskBase, gather)

        PypeScatteredTasks = pypeflow.task.PypeScatteredTasks

        @PypeScatteredTasks( inputDataObjs = {"inf":infileObj},
                             outputDataObjs = {"outf":outfileObj} )
        def test_fun(*argv, **kwargv):
            chunk_id = kwargv["chunk_id"]
            self = test_fun[chunk_id]
            assert self.inf._path == "/tmp/pypetest/%03d_test_in_1.txt" % chunk_id
            with open(  self.outf._path , "w") as f:
                in_f = open(self.inf.localFileName,"r")
                f.write("out:"+in_f.read())
                in_f.close()
            return self.inf._path

        assert len(test_fun.getTasks()) == nChunk 
        for i in range(nChunk):
            test_fun[i]()

        outfileObj.getGatherTask()()
        
    def test_pype_scattered_tasks_2(self):
        import os
        #os.system("rm -rf /tmp/pypetest/*")

        nChunk = 5 

        infileObj =\
        pypeflow.data.PypeSplittableLocalFile(
                      "splittablefile://localhost/tmp/pypetest/test_in_2.txt", 
                      nChunk = nChunk)

        with open(infileObj.localFileName, "w") as f:
            for i in range(nChunk):
                f.write("file%02d\n" % i)

        with open("/tmp/pypetest/scatter.sh", "w") as f:
            f.write("#!/bin/bash\n")
            f.write("for f in %s;" % " ".join( ["%03d" % i for i in range(nChunk)] )) 
            f.write('do if [ -e /tmp/pypetest/%f"_test_in.txt" ];\
                        then rm /tmp/pypetest/$f"_test_in.txt"; fi;\n')
            f.write("done\n")
            for i in range(nChunk):
                f.write("echo file%02d > /tmp/pypetest/%03d_test_in_2.txt\n" % (i, i))

        PypeShellTask = pypeflow.task.PypeShellTask
        PypeTask = pypeflow.task.PypeTask
        PypeTaskBase = pypeflow.task.PypeTaskBase
        infileObj.setScatterTask(PypeShellTask, PypeTaskBase, "/tmp/pypetest/scatter.sh")
        infileObj.getScatterTask()()
        
        def gather(*argv, **kwargv):
            inputObjs = sorted( kwargv["inputDataObjs"].items() )
            with open(kwargv["outputDataObjs"]["completeFile"].localFileName,"w") as outf:
                for k, subfile in inputObjs:
                    f = open(subfile.localFileName)
                    outf.write("out:"+f.read())
                    f.close()

        outfileObj =\
        pypeflow.data.PypeSplittableLocalFile(
                      "splittablefile://localhost/tmp/pypetest/test_out_2.txt", 
                      nChunk = nChunk)

        outfileObj.setGatherTask(PypeTask, PypeTaskBase, gather)

        PypeScatteredTasks = pypeflow.task.PypeScatteredTasks

        @PypeScatteredTasks( inputDataObjs = {"inf":infileObj},
                             outputDataObjs = {"outf":outfileObj},
                             comment="xyz")
        def test_fun_2(*argv, **kwargv):
            assert kwargv["comment"] == "xyz"
            chunk_id = kwargv["chunk_id"]
            self = test_fun_2[chunk_id]
            assert self.inf._path == "/tmp/pypetest/%03d_test_in_2.txt" % chunk_id
            with open(  self.outf._path , "w") as f:
                f.write("file%02d\n" % chunk_id)
            return self.inf._path

        assert len(test_fun_2.getTasks()) == nChunk 
        for i in range(nChunk):
            test_fun_2[i]()

        outfileObj.getGatherTask()()

    def test_pype_scattered_tasks_3(self):
        import os
        #os.system("rm -rf /tmp/pypetest/*")
        nChunk = 5 


        infileObj0 =\
        pypeflow.data.PypeLocalFile(
                      "file://localhost/tmp/pypetest/test_in_0.txt")
        with open(infileObj0.localFileName,"w") as f:
            f.write("prefix:")

        infileObj =\
        pypeflow.data.PypeSplittableLocalFile(
                      "splittablefile://localhost/tmp/pypetest/test_in_3.txt", 
                      nChunk = nChunk)

        with open(infileObj.localFileName, "w") as f:
            for i in range(nChunk):
                f.write("file%02d\n" % i)

        def scatter(*argv, **kwargv):
            outputObjs = sorted( kwargv["outputDataObjs"].items() )
            nOut = len(outputObjs)
            outputObjs = [ (o[0], o[1], open(o[1].localFileName, "w")) for o in outputObjs]
            with open(kwargv["inputDataObjs"]["completeFile"].localFileName,"r") as f:
                i = 0
                for l in f:
                    outf = outputObjs[i % nOut][2]
                    outf.write(l)
                    i += 1
            for o in outputObjs:
                o[2].close()

        PypeShellTask = pypeflow.task.PypeShellTask
        PypeTask = pypeflow.task.PypeTask
        PypeTaskBase = pypeflow.task.PypeTaskBase
        infileObj.setScatterTask(PypeTask, PypeTaskBase, scatter)
        infileObj.getScatterTask()()
        
        def gather(*argv, **kwargv):
            inputObjs = sorted( kwargv["inputDataObjs"].items() )
            with open(kwargv["outputDataObjs"]["completeFile"].localFileName,"w") as outf:
                for k, subfile in inputObjs:
                    f = open(subfile.localFileName)
                    outf.write(f.read())
                    f.close()

        outfileObj3 =\
        pypeflow.data.PypeSplittableLocalFile(
                      "splittablefile://localhost/tmp/pypetest/test_out_3.txt", 
                      nChunk = nChunk)

        outfileObj3.setGatherTask(PypeTask, PypeTaskBase, gather)

        PypeScatteredTasks = pypeflow.task.PypeScatteredTasks

        @PypeScatteredTasks( inputDataObjs = {"inf":infileObj, "prefix":infileObj0},
                             outputDataObjs = {"outf":outfileObj3} )
        def test_fun_3(*argv, **kwargv):
            chunk_id = kwargv["chunk_id"]
            self = test_fun_3[chunk_id]

            assert self.inf._path == "/tmp/pypetest/%03d_test_in_3.txt" % chunk_id
            with open( self.prefix.localFileName, "r") as f:
                prefix = f.read()

            with open( self.outf._path, "w") as f:
                in_f = open(self.inf.localFileName,"r")
                f.write(prefix + in_f.read())
                in_f.close()
            return self.inf._path

        assert len(test_fun_3.getTasks()) == nChunk 
        for i in range(nChunk):
            test_fun_3[i]()

        outfileObj3.getGatherTask()()
